"""Groups all time program related functionality. Time program is used when
:class:`~pymultimatic.model.mode.OperatingModes.AUTO` operation mode is
activated."""
import copy
from datetime import timedelta, datetime
from typing import List, Dict, Optional, Any

import re
import attr

from . import SettingMode


def _to_absolute_minutes(start_time: str) -> int:
    """Convert hh:mm to absolute minutes."""
    split = start_time.split(":")
    hour = int(split[0]) * 60
    minute = int(split[1])
    return hour + minute


# pylint: disable=too-few-public-methods
@attr.s
class TimePeriodSetting:
    """This is a period setting, defining what the
    :class:`~pymultimatic.model.component.Component` should do when on
    :class:`~pymultimatic.model.mode.OperatingModes.AUTO`.

    There is not end of the period, the end of the period is equal to the start
    time of the next period.

    Args:
        start_time (str): Start time of the period, format is HH:mm.
        target_temperature (float): Target temperature to reach during this
        time period, target temperature is only available for
            :class:`~pymultimatic.model.component.Room`.
        setting (SettingMode): The setting configured, see
            :class:`~pymultimatic.model.mode.SettingModes`, this is not
            available for a :class:`~pymultimatic.model.component.Room`.
    """

    start_time = attr.ib(type=str)
    target_temperature = attr.ib(type=Optional[float])
    setting = attr.ib(type=Optional[SettingMode])
    hour = attr.ib(type=int, init=False)
    """Hour at which the setting is starting."""
    minute = attr.ib(type=int, init=False)
    """Minute at which the setting is starting."""
    absolute_minutes = attr.ib(type=int, init=False)
    """Represents :attr:`start_time` in absolute minute. hours * 60 + minutes.
    This is more convenient to compare TimePeriodSetting using this."""

    def __attrs_post_init__(self) -> None:
        self.absolute_minutes = _to_absolute_minutes(self.start_time)
        self.hour = int(self.start_time.split(':')[0])
        self.minute = int(self.start_time.split(':')[1])

    # pylint: disable=unused-argument, no-self-use
    @start_time.validator
    def _validate_start_time(self, attribute: Any, value: Any) -> None:
        validator = re.compile('[0-9]{2}:[0-9]{2}')
        if not validator.match(value):
            raise ValueError(value)

    def __deepcopy__(self, memodict: Any = None) -> 'TimePeriodSetting':
        return TimePeriodSetting(self.start_time, self.target_temperature,
                                 self.setting)


# pylint: disable=too-few-public-methods
@attr.s
class TimeProgramDay:
    """This is a day, this is basically a list of :class:`TimePeriodSetting`.

    Args:
        settings (List[TimePeriodSetting]): List of periods for this day.
    """

    settings = attr.ib(type=List[TimePeriodSetting])


# pylint: disable=too-few-public-methods
@attr.s
class TimeProgram:
    """This is the full time program, a week, reflecting the configuration done
    through mobile app.

    Args:
        days (Dict[str, TimeProgramDay]): Days of the week.
    """

    days = attr.ib(type=Dict[str, TimeProgramDay])

    def get_for(self, search_date: datetime) -> TimePeriodSetting:
        """Get the corresponding :class:`TimePeriodSetting`
        for the given time.

        Args:
            search_date (datetime): Only the day, the hour and minute are used
                in order to get the right :class:`TimePeriodSetting`.

        Returns:
            TimePeriodSetting: The corresponding setting.
        """
        day = search_date.strftime("%A").lower()
        day_before = (search_date - timedelta(days=1)).strftime("%A").lower()
        time = str(search_date.hour) + ':' + str(search_date.minute)

        abs_minutes = _to_absolute_minutes(time)
        tp_day = self.days[day]
        tp_day_before = self.days[day_before]

        # if given hour:minute is before the first setting of the day,
        # get last setting of the previous day
        if abs_minutes < tp_day.settings[0].absolute_minutes:
            return copy.deepcopy(tp_day_before.settings[-1])

        idx: int = 0
        max_len: int = len(tp_day.settings)
        while idx < max_len and\
                abs_minutes > tp_day.settings[idx].absolute_minutes:
            idx += 1

        if not idx == max_len:
            # At this point, we went 1 step too far, so idx - 1
            return copy.deepcopy(tp_day.settings[idx - 1])

        # if no match a this point, it means search date is after the last
        # setting of the day
        return copy.deepcopy(tp_day.settings[-1])

    def get_next(self, search_date: datetime) -> TimePeriodSetting:
        """
        Get the next :class:`TimePeriodSetting` for the given time.
        So if there is a time program like this:

        * 5:00

        * 6:00

        And It's 5:30, it will return the second one

        Args:
            search_date (datetime): Only the day, the hour and minute are used
                in order to get the right :class:`TimePeriodSetting`.

        Returns:
            TimePeriodSetting: The corresponding setting.
        """
        day = search_date.strftime("%A").lower()
        day_after = (search_date + timedelta(days=1)).strftime("%A").lower()
        time = str(search_date.hour) + ':' + str(search_date.minute)

        abs_minutes = _to_absolute_minutes(time)
        tp_day = self.days[day]
        tp_day_after = self.days[day_after]

        # if given hour:minute is before the first setting of the day,
        # get the first setting of the day
        if abs_minutes < tp_day.settings[0].absolute_minutes:
            return copy.deepcopy(tp_day.settings[0])

        idx: int = 0
        max_len: int = len(tp_day.settings)
        while idx < max_len and \
                abs_minutes > tp_day.settings[idx].absolute_minutes:
            idx += 1

        if not idx == max_len:
            # At this point, we are on next setting, so it's fine
            return copy.deepcopy(tp_day.settings[idx])

        # if no match a this point, it means search date is after the last
        # setting of the day, so next one, is the first setting of the next day
        return copy.deepcopy(tp_day_after.settings[0])



"""Vaillant API Urls with placeholder and when needed.

All placeholders are resolved here except ``{serial_number}`` which is resolved
by :mod:`~pymultimatic.api.connector`
"""
from urllib import parse

_BASE = 'https://smart.vaillant.com/mobile/api/v4'

_BASE_AUTHENTICATE = _BASE + '/account/authentication/v1'
_AUTHENTICATE = _BASE_AUTHENTICATE + '/authenticate'
_NEW_TOKEN = _BASE_AUTHENTICATE + '/token/new'
_LOGOUT = _BASE_AUTHENTICATE + '/logout'

"""Facility details"""
_FACILITIES_LIST = _BASE + '/facilities'
_FACILITIES = _FACILITIES_LIST + '/{serial_number}'
_FACILITIES_DETAILS = _FACILITIES + '/system/v1/details'
_FACILITIES_STATUS = _FACILITIES + '/system/v1/status'
_FACILITIES_SETTINGS = _FACILITIES + '/storage'
_FACILITIES_DEFAULT_SETTINGS = _FACILITIES + '/storage/default'
_FACILITIES_INSTALLER_INFO = _FACILITIES + '/system/v1/installerinfo'

"""Rbr (Room by room)"""
_RBR_BASE = _FACILITIES + '/rbr/v1'
_RBR_INSTALLATION_STATUS = _RBR_BASE + '/installationStatus'
_RBR_UNDERFLOOR_HEATING_STATUS = _RBR_BASE + '/underfloorHeatingStatus'

"""Rooms"""
_ROOMS_LIST = _RBR_BASE + '/rooms'
_ROOM = _ROOMS_LIST + '/{room_index}'
_ROOM_CONFIGURATION = _ROOMS_LIST + '/{room_index}/configuration'
_ROOM_QUICK_VETO = _ROOM_CONFIGURATION + '/quickVeto'
_ROOM_TIMEPROGRAM = _ROOMS_LIST + '/{room_index}/timeprogram'
_ROOM_OPERATING_MODE = _ROOM_CONFIGURATION + '/operationMode'
_ROOM_CHILD_LOCK = _ROOM_CONFIGURATION + '/childLock'
_ROOM_NAME = _ROOM_CONFIGURATION + '/name'
_ROOM_DEVICE_NAME = _ROOM_CONFIGURATION + '/devices/{sgtin}/name'
_ROOM_TEMPERATURE_SETPOINT = _ROOM_CONFIGURATION + '/temperatureSetpoint'

"""Repeaters"""
_REPEATERS_LIST = _RBR_BASE + '/repeaters'
_REPEATER_DELETE = _REPEATERS_LIST + '/{sgtin}'
_REPEATER_SET_NAME = _REPEATERS_LIST + '/{sgtin}/name'

"""HVAC (heating, ventilation and Air-conditioning)"""
_HVAC = _FACILITIES + '/hvacstate/v1/overview'
_HVAC_REQUEST_UPDATE = _FACILITIES + '/hvacstate/v1/hvacMessages/update'

"""EMF (Embedded Metering Function)"""
_LIVE_REPORT = _FACILITIES + '/livereport/v1'
_LIVE_REPORT_DEVICE = _LIVE_REPORT + '/devices/{device_id}/reports/{report_id}'
_PHOTOVOLTAICS_REPORT = _FACILITIES + '/spine/v1/currentPVMeteringInfo'
_EMF_REPORT = _FACILITIES + '/emf/v1/devices'
_EMF_REPORT_DEVICE = _EMF_REPORT + '/{device_id}'

"""System control"""
_SYSTEM = _FACILITIES + '/systemcontrol/v1'
_SYSTEM_CONFIGURATION = _SYSTEM + '/configuration'
_SYSTEM_STATUS = _SYSTEM + '/status'
_SYSTEM_DATETIME = _SYSTEM_STATUS + '/datetime'
_SYSTEM_PARAMETERS = _SYSTEM + '/parameters'
_SYSTEM_QUICK_MODE = _SYSTEM_CONFIGURATION + '/quickmode'
_SYSTEM_HOLIDAY_MODE = _SYSTEM_CONFIGURATION + '/holidaymode'

"""DHW (Domestic Hot Water)"""
_DHW = _SYSTEM + '/dhw/{dhw_id}'

"""Circulation"""
_CIRCULATION = _DHW + '/circulation'
_CIRCULATION_CONFIGURATION = _CIRCULATION + '/configuration'
_CIRCULATION_TIMEPROGRAM = _CIRCULATION_CONFIGURATION + '/timeprogram'

"""Hot water"""
_HOT_WATER = _DHW + '/hotwater'
_HOT_WATER_CONFIGURATION = _HOT_WATER + '/configuration'
_HOT_WATER_TIMEPROGRAM = _HOT_WATER_CONFIGURATION + '/timeprogram'
_HOT_WATER_OPERATING_MODE = _HOT_WATER_CONFIGURATION + '/operation_mode'
_HOT_WATER_TEMPERATURE_SETPOINT = _HOT_WATER_CONFIGURATION + \
                                  '/temperature_setpoint'

"""Ventilation"""
_VENTILATION = _SYSTEM + '/ventilation/{ventilation_id}'
_VENTILATION_CONFIGURATION = _VENTILATION + '/fan/configuration'
_VENTILATION_TIMEPROGRAM = _VENTILATION_CONFIGURATION + '/timeprogram'
_VENTILATION_DAY_LEVEL = _VENTILATION_CONFIGURATION + '/day_level'
_VENTILATION_NIGHT_LEVEL = _VENTILATION_CONFIGURATION + '/night_level'
_VENTILATION_OPERATING_MODE = _VENTILATION_CONFIGURATION + '/operation_mode'

"""Zones"""
_ZONES_LIST = _SYSTEM + '/zones'
_ZONE = _ZONES_LIST + '/{zone_id}'
_ZONE_CONFIGURATION = _ZONE + '/configuration'
_ZONE_NAME = _ZONE_CONFIGURATION + '/name'
_ZONE_QUICK_VETO = _ZONE_CONFIGURATION + '/quick_veto'

"""Zone heating"""
_ZONE_HEATING_CONFIGURATION = _ZONE + '/heating/configuration'
_ZONE_HEATING_TIMEPROGRAM = _ZONE + '/heating/timeprogram'
_ZONE_HEATING_MODE = _ZONE_HEATING_CONFIGURATION + '/mode'
_ZONE_HEATING_SETPOINT_TEMPERATURE = _ZONE_HEATING_CONFIGURATION + \
                                     '/setpoint_temperature'
_ZONE_HEATING_SETBACK_TEMPERATURE = _ZONE_HEATING_CONFIGURATION + \
                                    '/setback_temperature'

"""Zone cooling"""
_ZONE_COOLING_CONFIGURATION = _ZONE + '/cooling/configuration'
_ZONE_COOLING_TIMEPROGRAM = _ZONE + '/cooling/timeprogram'
_ZONE_COOLING_MODE = _ZONE_COOLING_CONFIGURATION + '/mode'
_ZONE_COOLING_SETPOINT_TEMPERATURE = _ZONE_COOLING_CONFIGURATION + \
                                     '/setpoint_temperature'
_ZONE_COOLING_MANUAL_SETPOINT_TEMPERATURE = \
    _ZONE_COOLING_CONFIGURATION + '/manual_mode_cooling_temperature_setpoint'


def new_token() -> str:
    """Url to request a new token."""
    return _NEW_TOKEN


def authenticate() -> str:
    """Url to authenticate the user and receive cookies."""
    return _AUTHENTICATE


def logout() -> str:
    """Url to logout from the API, cookies are invalidated."""
    return _LOGOUT


def facilities_list() -> str:
    """Url to get the list of serial numbers of the facilities (and some other
    properties).

    Note:
        For now, the connector only handle one serial number.
    """
    return _FACILITIES_LIST


def rbr_underfloor_heating_status() -> str:
    """Url to check if underfloor heating is installed or not."""
    return _RBR_UNDERFLOOR_HEATING_STATUS.format(
        serial_number='{serial_number}')


def rbr_installation_status() -> str:
    """Url to check the room by room installation status."""
    return _RBR_INSTALLATION_STATUS.format(
        serial_number='{serial_number}')


def rooms() -> str:
    """Url to get the list of :class:`~pymultimatic.model.component.Room`."""
    return _ROOMS_LIST.format(serial_number='{serial_number}')


def room(room_index: str) -> str:
    """Url to get specific room details (configuration, timeprogram). Or to
    delete a :class:`~pymultimatic.model.component.Room`.
    """
    return _ROOM.format(serial_number='{serial_number}',
                        room_index=room_index)


def room_configuration(room_index: str) -> str:
    """Url to get configuration for a
    :class:`~pymultimatic.model.component.Room` (name, temperature,
    target temperature, etc.).
    """
    return _ROOM_CONFIGURATION.format(serial_number='{serial_number}',
                                      room_index=room_index)


def room_quick_veto(room_index: str) -> str:
    """Url to handle :class:`~pymultimatic.model.mode.QuickVeto` for a
    :class:`~pymultimatic.model.component.Room`.
    """
    return _ROOM_QUICK_VETO.format(serial_number='{serial_number}',
                                   room_index=room_index)


def room_operating_mode(room_index: str) -> str:
    """Url to set operating for a :class:`~pymultimatic.model.component.Room`.
    """
    return _ROOM_OPERATING_MODE.format(
        serial_number='{serial_number}', room_index=room_index)


def room_timeprogram(room_index: str) -> str:
    """Url to get/update configuration for a
    class:`~pymultimatic.model.component.Room`. (name, temperature,
    target temperature, etc.).
    """
    return _ROOM_TIMEPROGRAM.format(serial_number='{serial_number}',
                                    room_index=room_index)


def room_child_lock(room_index: str) -> str:
    """Url to handle child lock for all
    :class:`~pymultimatic.model.component.Device` in a
    :class:`~pymultimatic.model.component.Room`.
    """
    return _ROOM_CHILD_LOCK.format(serial_number='{serial_number}',
                                   room_index=room_index)


def room_name(room_index: str) -> str:
    """Set :class:`~pymultimatic.model.component.Room` name."""
    return _ROOM_NAME.format(serial_number='{serial_number}',
                             room_index=room_index)


def room_device_name(room_index: str, sgtin: str) -> str:
    """Set :class:`~pymultimatic.model.component.Device` name."""
    return _ROOM_DEVICE_NAME.format(serial_number='{serial_number}',
                                    room_index=room_index,
                                    sgtin=sgtin)


def room_temperature_setpoint(room_index: str) -> str:
    """Url to handle target temperature for a
    :class:`~pymultimatic.model.component.Room`.
    """
    return _ROOM_TEMPERATURE_SETPOINT.format(
        serial_number='{serial_number}', room_index=room_index)


def repeaters() -> str:
    """Url to get list of repeaters"""
    return _REPEATERS_LIST.format(serial_number='{serial_number}')


def delete_repeater(sgtin: str) -> str:
    """Url to delete a repeater."""
    return _REPEATER_DELETE.format(serial_number='{serial_number}',
                                   sgtin=sgtin)


def repeater_name(sgtin: str) -> str:
    """Url to set repeater's name."""
    return _REPEATER_SET_NAME.format(serial_number='{serial_number}',
                                     sgtin=sgtin)


def hvac() -> str:
    """Url of the hvac overview."""
    return _HVAC.format(serial_number='{serial_number}')


def hvac_update() -> str:
    """Url to request an hvac update."""
    return _HVAC_REQUEST_UPDATE.format(
        serial_number='{serial_number}')


def live_report() -> str:
    """Url to get live report data (current boiler water temperature, current
    hot water temperature, etc.)."""
    return _LIVE_REPORT.format(serial_number='{serial_number}')


def live_report_device(device_id: str, report_id: str) -> str:
    """
    Url to get live report for specific device
    """
    return _LIVE_REPORT_DEVICE.format(serial_number='{serial_number}',
                                      device_id=device_id,
                                      report_id=report_id)


def photovoltaics() -> str:
    """Url to get photovoltaics data."""
    return _PHOTOVOLTAICS_REPORT.format(
        serial_number='{serial_number}')


def emf_report() -> str:
    """Url to get emf (Embedded Metering Function) report."""
    return _EMF_REPORT.format(serial_number='{serial_number}')


# pylint: disable=too-many-arguments
def emf_report_device(device_id: str, energy_type: str, function: str,
                      time_range: str, start: str, offset: str) -> str:
    """Url to get emf (Embedded Metering Function) report for a specific
    device."""
    url = _EMF_REPORT_DEVICE.format(serial_number='{serial_number}',
                                    device_id=device_id)

    query_params = {
        'energy_type': energy_type,
        'function': function,
        'timeRange': time_range,
        'start': start,
        'offset': offset,
    }

    return '{}?{}'.format(url, parse.urlencode(query_params))


def facilities_details() -> str:
    """Url to get facility detail."""
    return _FACILITIES_DETAILS.format(serial_number='{serial_number}')


def facilities_status() -> str:
    """Url to get facility status."""
    return _FACILITIES_STATUS.format(serial_number='{serial_number}')


def facilities_settings() -> str:
    """Url to get facility settings."""
    return _FACILITIES_SETTINGS.format(
        serial_number='{serial_number}')


def facilities_default_settings() -> str:
    """
    Url to get facility default settings
    """
    return _FACILITIES_DEFAULT_SETTINGS.format(
        serial_number='{serial_number}')


def facilities_installer_info() -> str:
    """Url to get facility default settings."""
    return _FACILITIES_INSTALLER_INFO.format(
        serial_number='{serial_number}')


def system() -> str:
    """Url to get full :class:`~pymultimatic.model.system.System` (zones, dhw,
    ventilation, holiday mode, etc.) except
    :class:`~pymultimatic.model.component.Room`.
    """
    return _SYSTEM.format(serial_number='{serial_number}')


def system_configuration() -> str:
    """Url to get system configuration (holiday mode, quick mode etc.)."""
    return _SYSTEM_CONFIGURATION.format(
        serial_number='{serial_number}')


def system_status() -> str:
    """Url to get outdoor temperature and datetime."""
    return _SYSTEM_STATUS.format(serial_number='{serial_number}')


def system_datetime() -> str:
    """Url to set datetime."""
    return _SYSTEM_DATETIME.format(serial_number='{serial_number}')


def system_parameters() -> str:
    """Url to get system parameters."""
    return _SYSTEM_PARAMETERS.format(serial_number='{serial_number}')


def system_quickmode() -> str:
    """Url to get system :class:`~pymultimatic.model.mode.QuickMode`."""
    return _SYSTEM_QUICK_MODE.format(serial_number='{serial_number}')


def system_holiday_mode() -> str:
    """Url to get system :class:`~pymultimatic.model.mode.HolidayMode`."""
    return _SYSTEM_HOLIDAY_MODE.format(
        serial_number='{serial_number}')


def dhw(dhw_id: str) -> str:
    """Url to get domestic hot water
    (:class:`~pymultimatic.model.component.HotWater` and
    :class:`~pymultimatic.model.component.Circulation`).
    """
    return _DHW.format(serial_number='{serial_number}', dhw_id=dhw_id)


def circulation(dhw_id: str) -> str:
    """Url to get :class:`~pymultimatic.model.component.Circulation` details.
    """
    return _CIRCULATION.format(serial_number='{serial_number}', dhw_id=dhw_id)


def circulation_configuration(dhw_id: str) -> str:
    """Url to handle :class:`~pymultimatic.model.component.Circulation`
    configuration.
    """
    return _CIRCULATION_CONFIGURATION.format(
        serial_number='{serial_number}', dhw_id=dhw_id)


def circulation_timeprogram(dhw_id: str) -> str:
    """Url to handle :class:`~pymultimatic.model.component.Circulation`
    :class:`~pymultimatic.model.timeprogram.TimeProgram`.
    """
    return _CIRCULATION_TIMEPROGRAM.format(
        serial_number='{serial_number}', dhw_id=dhw_id)


def hot_water(dhw_id: str) -> str:
    """Url to get :class:`~pymultimatic.model.component.HotWater` detail."""
    return _HOT_WATER.format(serial_number='{serial_number}',
                             dhw_id=dhw_id)


def hot_water_configuration(dhw_id: str) -> str:
    """Url to handle :class:`~pymultimatic.model.component.HotWater`
    configuration.
    """
    return _HOT_WATER_CONFIGURATION.format(
        serial_number='{serial_number}', dhw_id=dhw_id)


def hot_water_timeprogram(dhw_id: str) -> str:
    """Url to handle :class:`~pymultimatic.model.component.HotWater`
    :class:`~pymultimatic.model.timeprogram.TimeProgram`.
    """
    return _HOT_WATER_TIMEPROGRAM.format(
        serial_number='{serial_number}', dhw_id=dhw_id)


def hot_water_operating_mode(dhw_id: str) -> str:
    """Url to set :class:`~pymultimatic.model.component.HotWater`
    operating mode, only if it's not a quick action.
    """
    return _HOT_WATER_OPERATING_MODE.format(
        serial_number='{serial_number}', dhw_id=dhw_id)


def hot_water_temperature_setpoint(dhw_id: str) -> str:
    """Url to set :class:`~pymultimatic.model.component.HotWater`
    temperature setpoint.
    """
    return _HOT_WATER_TEMPERATURE_SETPOINT.format(
        serial_number='{serial_number}', dhw_id=dhw_id)


def ventilation(ventilation_id: str) -> str:
    """Url to get ventilation details."""
    return _VENTILATION.format(serial_number='{serial_number}',
                               ventilation_id=ventilation_id)


def ventilation_configuration(ventilation_id: str) -> str:
    """Url to get ventilation configuration."""
    return _VENTILATION_CONFIGURATION.format(
        serial_number='{serial_number}', ventilation_id=ventilation_id)


def ventilation_timeprogram(ventilation_id: str) -> str:
    """Url to get ventilation timeprogram."""
    return _VENTILATION_TIMEPROGRAM.format(
        serial_number='{serial_number}', ventilation_id=ventilation_id)


def set_ventilation_day_level(ventilation_id: str) -> str:
    """Url to set ventilation day level."""
    return _VENTILATION_DAY_LEVEL.format(
        serial_number='{serial_number}', ventilation_id=ventilation_id)


def set_ventilation_night_level(ventilation_id: str) -> str:
    """
    Url to set ventilation night level
    """
    return _VENTILATION_NIGHT_LEVEL.format(
        serial_number='{serial_number}', ventilation_id=ventilation_id)


def set_ventilation_operating_mode(ventilation_id: str) -> str:
    """Url to set ventilation operating mode."""
    return _VENTILATION_OPERATING_MODE.format(
        serial_number='{serial_number}', ventilation_id=ventilation_id)


def zones() -> str:
    """Url to get :class:`~pymultimatic.model.component.Zone`."""
    return _ZONES_LIST.format(serial_number='{serial_number}')


def zone(zone_id: str) -> str:
    """Url to get a specific :class:`~pymultimatic.model.component.Zone`."""
    return _ZONE.format(serial_number='{serial_number}',
                        zone_id=zone_id)


def zone_configuration(zone_id: str) -> str:
    """Url to get a specific :class:`~pymultimatic.model.component.Zone`
    configuration.
    """
    return _ZONE_CONFIGURATION.format(serial_number='{serial_number}',
                                      zone_id=zone_id)


def zone_name(zone_id: str) -> str:
    """Url to set :class:`~pymultimatic.model.component.Zone` name."""
    return _ZONE_NAME.format(serial_number='{serial_number}',
                             zone_id=zone_id)


def zone_quick_veto(zone_id: str) -> str:
    """Url to get :class:`~pymultimatic.model.mode.QuickVeto` for a
    :class:`~pymultimatic.model.component.Zone`.
    """
    return _ZONE_QUICK_VETO.format(serial_number='{serial_number}',
                                   zone_id=zone_id)


def zone_heating_configuration(zone_id: str) -> str:
    """Url to get :class:`~pymultimatic.model.component.Zone` heating
    configuration.
    """
    return _ZONE_HEATING_CONFIGURATION.format(
        serial_number='{serial_number}', zone_id=zone_id)


def zone_heating_timeprogram(zone_id: str) -> str:
    """Url to get a :class:`~pymultimatic.model.component.Zone` heating
    :class:`~pymultimatic.model.timeprogram.TimeProgram`.
    """
    return _ZONE_HEATING_TIMEPROGRAM.format(
        serial_number='{serial_number}', zone_id=zone_id)


def zone_heating_mode(zone_id: str) -> str:
    """Url to get a :class:`~pymultimatic.model.component.Zone` heating mode.
    """
    return _ZONE_HEATING_MODE.format(serial_number='{serial_number}',
                                     zone_id=zone_id)


def zone_heating_setpoint_temperature(zone_id: str) -> str:
    """Url to set a :class:`~pymultimatic.model.component.Zone` setpoint
    temperature.
    """
    return _ZONE_HEATING_SETPOINT_TEMPERATURE.format(
        serial_number='{serial_number}', zone_id=zone_id)


def zone_heating_setback_temperature(zone_id: str) -> str:
    """Url to set a :class:`~pymultimatic.model.component.Zone` setback
    temperature.
    """
    return _ZONE_HEATING_SETBACK_TEMPERATURE.format(
        serial_number='{serial_number}', zone_id=zone_id)


def zone_cooling_configuration(zone_id: str) -> str:
    """Url to get a :class:`~pymultimatic.model.component.Zone` cooling
    configuration.
    """
    return _ZONE_COOLING_CONFIGURATION.format(
        serial_number='{serial_number}', zone_id=zone_id)


def zone_cooling_timeprogram(zone_id: str) -> str:
    """Url to get :class:`~pymultimatic.model.component.Zone` cooling
    timeprogram.
    """
    return _ZONE_COOLING_TIMEPROGRAM.format(
        serial_number='{serial_number}', zone_id=zone_id)


def zone_cooling_mode(zone_id: str) -> str:
    """Url to set a :class:`~pymultimatic.model.component.Zone` cooling mode.
    """
    return _ZONE_COOLING_MODE.format(serial_number='{serial_number}',
                                     zone_id=zone_id)


def zone_cooling_setpoint_temperature(zone_id: str) -> str:
    """Url to set the cooling temperature setpoint for a
    :class:`~pymultimatic.model.component.Zone`.
    """
    return _ZONE_COOLING_SETPOINT_TEMPERATURE.format(
        serial_number='{serial_number}', zone_id=zone_id)


def zone_cooling_manual_setpoint_temperature(zone_id: str) -> str:
    """Url to set manual cooling setpoint temperature for a
    :class:`~pymultimatic.model.component.Zone`.
    """
    return _ZONE_COOLING_MANUAL_SETPOINT_TEMPERATURE.format(
        serial_number='{serial_number}', zone_id=zone_id)



"""Grouping status"""
from datetime import datetime
import attr


# pylint: disable=too-few-public-methods
@attr.s
class Error:
    """Errors coming from your system.

    Args:
        device_name (str): Name of the device from where the error is coming.
        title (str): Short description of the error.
        status_code (str): Code of the error.
        description (str): Long description of the error.
        timestamp (datetime): When errors occurred.
    """

    device_name = attr.ib(type=str)
    title = attr.ib(type=str)
    status_code = attr.ib(type=str)
    description = attr.ib(type=str)
    timestamp = attr.ib(type=datetime)


# pylint: disable=too-few-public-methods
@attr.s
class BoilerStatus(Error):
    """Status of the boiler. This is sent with an error format, but in this
    case, it's more like a status.

    Args:
        hint (str): Directly coming from the API, most of the time the value is
            ``...``
    """

    hint = attr.ib(type=str)

    @property
    def is_error(self) -> bool:
        """bool: Checks if there is an error at boiler side."""
        return self.status_code is not None \
            and (self.status_code.startswith('F') or self.status_code == 'con')


@attr.s
class SystemStatus:
    """Status of the system."""

    online_status = attr.ib(type=str)
    update_status = attr.ib(type=str)

    @property
    def is_online(self) -> bool:
        """bool: Checks if the system is connected to the internet."""
        return self.online_status == 'ONLINE'

    @property
    def is_up_to_date(self) -> bool:
        """bool: Checks if the system is up to date."""
        return self.update_status == 'UPDATE_NOT_PENDING'



"""Components of vaillant system"""
import abc
from typing import Optional, List
from datetime import datetime
import attr

from . import Mode, TimeProgram, QuickVeto, ActiveMode, OperatingMode, \
    OperatingModes, SettingModes, constants


# pylint: disable=too-few-public-methods
@attr.s
class Component:
    """This is a common class for components in the system.
    A component can be :class:`Room`, :class:`Zone`, :class:`HotWater` or
    :class:`Circulation`.

    It allows you to get the :class:`~pymultimatic.model.mode.ActiveMode`.

    Note:
        :class:`active_mode` should not be used directly through the component,
        actually, the component is only aware of itself, so in case of a
        :class:`~pymultimatic.model.mode.QuickMode` is running on, you may
        not receive the active mode that is really applied. You should use
        :class:`~pymultimatic.model.system.System` for that.

    Args:
        id (str): Id of the component, this is used to get data from the API.
        name (str): Name of the component, this is set by the user through the
            mobile app or via the VRC700.
        time_program (TimeProgram): Used for time based operating mode,
            :class:`~pymultimatic.model.mode.OperatingModes.AUTO`
        current_temperature (float): Current temperature for the component,
            this is `None` for :class:`Circulation`
        target_temperature (float): Temperature the component is trying to
            reach, also `None` for :class:`Circulation`
        operating_mode (OperatingMode): Configured operating mode.
        quick_veto (QuickVeto): Will be populated if there is a
            :class:`~pymultimatic.model.mode.QuickVeto` running on. Always
            `None` for :class:`Circulation` and :class:`HotWater`.
    """

    # pylint: disable=invalid-name
    id = attr.ib(type=str)
    name = attr.ib(type=Optional[str])
    time_program = attr.ib(type=TimeProgram)
    current_temperature = attr.ib(type=Optional[float])
    target_temperature = attr.ib(type=Optional[float])
    operating_mode = attr.ib(type=OperatingMode)
    quick_veto = attr.ib(type=Optional[QuickVeto])

    @property
    def active_mode(self) -> ActiveMode:
        """ActiveMode: Get the :class:`~pymultimatic.model.mode.ActiveMode`
        for this component. All operating modes and quick veto are handled,
        **but not quick mode.**

        Note:
        :class:`Component.active_mode` should not be used directly through the
        component, actually, the component is only aware of itself, so in case
        of a :class:`~pymultimatic.model.mode.QuickMode` is running on, you
        may not receive the active mode that is really applied. You should use
        :class:`~pymultimatic.model.system.System` for that.
        """

        if self.quick_veto:
            return ActiveMode(self.quick_veto.target_temperature,
                              OperatingModes.QUICK_VETO)

        if self.operating_mode == OperatingModes.AUTO:
            setting = self.time_program.get_for(datetime.now())
            if setting.target_temperature:
                return ActiveMode(setting.target_temperature,
                                  OperatingModes.AUTO, setting.setting)

        return self._get_specific_active_mode()

    @abc.abstractmethod
    def _get_specific_active_mode(self) -> ActiveMode:
        """Get the specific active mode for a component."""


# pylint: disable=too-few-public-methods
@attr.s
class Circulation(Component):
    """This is representing the circulation from the system.
    This is a bit special component since there is no
    `current_temperature`, `target_temperature` nor `quick_veto`.
    """

    MODES: List[Mode] = [OperatingModes.ON, OperatingModes.OFF,
                         OperatingModes.AUTO]
    """List of mode that are applicable to the circulation component."""

    current_temperature = attr.ib(default=None, init=False)
    target_temperature = attr.ib(default=None, init=False)
    quick_veto = attr.ib(default=None, init=False)

    def _get_specific_active_mode(self) -> ActiveMode:
        """Gets specific active mode for a component."""
        if self.operating_mode == OperatingModes.AUTO:
            setting = self.time_program.get_for(datetime.now())
            mode = ActiveMode(None, OperatingModes.AUTO, setting.setting)
        else:
            mode = ActiveMode(None, self.operating_mode)
        return mode


# pylint: disable=too-few-public-methods
@attr.s
class HotWater(Component):
    """This is representing the hot water from the system.

    There is no `quick_veto` available for this component.
    """

    MODES = [OperatingModes.ON, OperatingModes.OFF, OperatingModes.AUTO]
    """List of mode that are applicable to the hot water component."""

    MIN_TARGET_TEMP = 35
    """Min `target temperature` for the hot water."""

    MAX_TARGET_TEMP = 70
    """Max `target temperature` for the hot water."""

    quick_veto = attr.ib(default=None, init=False)

    def _get_specific_active_mode(self) -> ActiveMode:
        """Gets specific active mode for a component."""
        if self.operating_mode == OperatingModes.AUTO:
            setting = self.time_program.get_for(datetime.now())

            if setting.setting == SettingModes.ON:
                mode = ActiveMode(self.target_temperature, OperatingModes.AUTO,
                                  setting.setting)
            else:
                mode = ActiveMode(constants.FROST_PROTECTION_TEMP,
                                  OperatingModes.AUTO, setting.setting)

        elif self.operating_mode == OperatingModes.ON:
            mode = ActiveMode(self.target_temperature, OperatingModes.ON)
        else:  # MODE_OFF
            mode = ActiveMode(constants.FROST_PROTECTION_TEMP,
                              OperatingModes.OFF)

        return mode


# pylint: disable=too-few-public-methods
@attr.s
class Device:
    """This is a physical device inside a :class:`Room`. It can be a VR50
    VR51 or VR52.

    Args:
        name (str): Name of the device, this is set by the mobile app.
        sgtin (str): The is the serial number of the device.
        device_type (str): Device type (VR51 = 'VALVE').
        battery_low (bool): Indicate if the device is running low battery.
        radio_out_of_reach (bool): Indicate if the device is connected to the
            system.
    """

    name = attr.ib(type=str)
    sgtin = attr.ib(type=str)
    device_type = attr.ib(type=str)
    battery_low = attr.ib(type=bool)
    radio_out_of_reach = attr.ib(type=bool)


# pylint: disable=too-few-public-methods
@attr.s
class Room(Component):
    """This is representing a room from the system.

    Note:
        The boost function - *the boost function is activated by pressing the
        temperature selector. This means that the heating valves on the
        assigned radiator thermostats are immediately opened 80% for 5
        minutes* - is not reflected into the API.

    Args:
        child_lock (bool): Indicate if the device is locked, meaning you cannot
            use the buttons on the device.
        window_open (bool): Indicate if a window is open in the room. Of
            course, this is data interpretation from vaillant API.
        devices (List[Device]): List of :class:`Device` inside the room.
        humidity (float): In case of a VR51 is inside the room, you can get
            humidity through the API.
    """

    MODES = [OperatingModes.OFF, OperatingModes.MANUAL, OperatingModes.AUTO,
             OperatingModes.QUICK_VETO]
    """List of mode that are applicable to rooms component."""

    MIN_TARGET_TEMP = constants.FROST_PROTECTION_TEMP
    """Min `target temperature` that can be apply to a room."""

    MAX_TARGET_TEMP = constants.THERMOSTAT_MAX_TEMP
    """Max `target temperature` that can be apply to a room."""

    child_lock = attr.ib(type=bool)
    window_open = attr.ib(type=bool)
    devices = attr.ib(type=List[Device])
    humidity = attr.ib(type=Optional[float], default=None)

    def _get_specific_active_mode(self) -> ActiveMode:
        """Gets specific active mode for a component."""
        if self.operating_mode == OperatingModes.OFF:
            mode = ActiveMode(self.MIN_TARGET_TEMP, OperatingModes.OFF)
        else:  # MODE_MANUAL
            mode = ActiveMode(self.target_temperature, OperatingModes.MANUAL)

        return mode

# pylint: disable=too-few-public-methods
@attr.s
class Zone(Component):
    """This is representing a zone from the system.

    Note:
        If `rbr` (Room By Room) is True, the zone itself doesn't mean anything
        anymore, it means rooms are 'controlling' the zone.

    Args:
        target_min_temperature (float): This is the set back temperature, so
            if the temperature goes below that value, the zone should heat.
        active_function (str): Indicate what the zone is doing (basically
            'HEATING' or 'STANDBY').
        rbr (bool): Room By Room, means the zone is controlled by ambisense.
    """

    MODES = [OperatingModes.AUTO, OperatingModes.OFF, OperatingModes.DAY,
             OperatingModes.NIGHT, OperatingModes.QUICK_VETO]
    """List of mode that are applicable to zones component."""

    MIN_TARGET_TEMP = constants.FROST_PROTECTION_TEMP
    """Min temperature that can be apply to a zone."""

    MAX_TARGET_TEMP = constants.THERMOSTAT_MAX_TEMP
    """Max temperature that can be apply to a zone."""

    target_min_temperature = attr.ib(type=float)
    active_function = attr.ib(type=str)
    rbr = attr.ib(type=bool)

    def _get_specific_active_mode(self) -> ActiveMode:
        """Gets specific active mode for a component."""
        if self.operating_mode == OperatingModes.AUTO:
            setting = self.time_program.get_for(datetime.now())

            if setting.setting == SettingModes.DAY:
                mode = ActiveMode(self.target_temperature, OperatingModes.AUTO,
                                  setting.setting)
            else:
                mode = ActiveMode(self.target_min_temperature,
                                  OperatingModes.AUTO, setting.setting)
        elif self.operating_mode == OperatingModes.OFF:
            mode = ActiveMode(self.MIN_TARGET_TEMP, OperatingModes.OFF)
        elif self.operating_mode == OperatingModes.DAY:
            mode = ActiveMode(self.target_temperature, OperatingModes.DAY)
        else:  # MODE_NIGHT
            mode = ActiveMode(self.target_min_temperature,
                              OperatingModes.NIGHT)

        return mode



"""Errors coming from the API."""
from typing import Optional, Any

from requests import Response

import attr


@attr.s
class ApiError(Exception):
    """This exception is thrown when a communication error occurs with the
    vaillant API."""

    message = attr.ib(type=str)
    response = attr.ib(type=Optional[Response])
    payload = attr.ib(type=Any, default=None)



"""contains some default parameters of the application."""

FILES_PATH = '~/.pymultimatic'
"""Default file path for cookies and serial."""

COOKIE_FILE_NAME = '.cookies'
"""Default cookie file name."""

SERIAL_NUMBER_FILE_NAME = '.serial'
"""Default serial file name."""

SMARTPHONE_ID = 'pymultiMATIC'
"""Default smartphone id."""



"""Vaillant operation modes."""
from datetime import date
from typing import Optional, List
import attr


# pylint: disable=too-few-public-methods
@attr.s
class Mode:
    """This is the base class for modes, it groups :class:`QuickVeto`,
    :class:`OperatingMode`, :class:`QuickMode` and :class:`SettingMode`.

    Args:
        name (str): The name of the mode.
    """

    name = attr.ib(type=str)


@attr.s(frozen=True)
class OperatingMode(Mode):
    """Represents the operating mode of a
    :class:`~pymultimatic.model.component.Component`"""


class OperatingModes:
    """Groups all :class:`OperatingMode` together."""

    AUTO = OperatingMode('AUTO')
    """The automatic mode controls the heating circuit in accordance with the
    set desired temperature and the set time slots.
    """

    DAY = OperatingMode('DAY')
    """The Day operating mode brings the heating circuit to the desired Day
    temperature set, without taking time periods into account.
    """

    NIGHT = OperatingMode('NIGHT')
    """The Set-back (Night) operating mode brings the heating circuit to the
    desired Set-back temperature set, without taking time periods into
    consideration.
    """

    ON = OperatingMode('ON')
    """The operating mode "On" applies to
    :class:`~pymultimatic.model.component.Hotwater` and
    :class:`~pymultimatic.model.component.Circulation`, it means the
    :class:`~pymultimatic.model.component.Component` is running on.
    """

    OFF = OperatingMode('OFF')
    """The operating mode "Off" switches the corresponding function off. The
    frost protection function is activated.
    """

    MANUAL = OperatingMode('MANUAL')
    """The manual operating mode is used by a
    :class:`~pymultimatic.model.component.Room` to set manually the target
    temperature and keep it until next manual change.
    """

    QUICK_VETO = OperatingMode('QUICK_VETO')
    """The Quick veto function can be used to make one-off changes to the
    temperature. This function sets the desired target temperature, which may
    differ from that configured in the
    :class:`~pymultimatic.model.timeprogram.TimeProgram` currently being used
    . The newly set temperature will be applied for a certain period of
    time."""

    _VALUES = {opm.name: opm for opm in [AUTO, DAY, NIGHT, ON, OFF, MANUAL,
                                         QUICK_VETO]}

    @classmethod
    def get(cls, name: str) -> OperatingMode:
        """Get :class:`OperatingMode` by name.

        Args:
            name (str): Name of the operating mode

        Returns:
            The corresponding operating mode or None
        """
        return cls._VALUES[name]


# pylint: disable=too-few-public-methods
@attr.s
class QuickVeto(Mode):
    """Represents a quick veto which can be applied to a
    :class:`~pymultimatic.model.component.Zone` or a
    :class:`~pymultimatic.model.component.Room`.

    For a :class:`~pymultimatic.model.component.Room`, quick veto duration is
    customizable and it's possible to get the remaining duration through the
    API. The max value is 1440 minutes (1 day)

    For a :class:`~pymultimatic.model.component.Zone`, quick veto duration is
    **not** customizable (default duration is 6 hours) and the API returns
    always *0*.

    Args:
        remaining_duration (int): remaining duration in minutes
            before the quick veto expires. Since remaining duration is not
            available for a :class:`~pymultimatic.model.component.Zone`,
            value will be `None`.
        target_temperature (float): target temperature defined by the quick
            veto.
    """

    name = attr.ib(default=OperatingModes.QUICK_VETO.name, init=False)
    remaining_duration = attr.ib(type=Optional[int])
    target_temperature = attr.ib(type=float)

    @remaining_duration.validator
    # pylint: disable=no-self-use
    def _remaining_duration_validator(self, attribute: str,
                                      value: Optional[int]) -> None:
        if value is not None and value > 1440:
            raise ValueError('{} with value {} is not valid'
                             .format(attribute, value))


@attr.s(frozen=True)
class SettingMode(Mode):
    """This is the setting which is configured in
    :class:`~pymultimatic.model.timeprogram.TimePeriodSetting`."""


class SettingModes:
    """Groups all :class:`SettingMode` together."""

    ON = SettingMode('ON')
    """Within the time period,
    :class:`~pymultimatic.model.component.HotWater` generation is switched on
    and maintains the hot water in the domestic hot water cylinder at the
    preset temperature. This also means the
    :class:`~pymultimatic.model.component.Circulation` is switched off.
    """

    OFF = SettingMode('OFF')
    """Outside the time period,
    :class:`~pymultimatic.model.component.HotWater` generation (and
    :class:`~pymultimatic.model.component.Circulation`) is switched off.
    """

    DAY = SettingMode('DAY')
    """Within the time periods, the controller brings the
    :class:`~pymultimatic.model.component.Zone` temperature to the `DAY`
    temp. heating that has been set (Comfort mode).
    """

    NIGHT = SettingMode('NIGHT')
    """Outside of the time period, the controller brings the
    :class:`~pymultimatic.model.component.Zone` temperature to the Set-back
    temp. heating (`NIGHT`) that has been set.
    """

    _VALUES = {sm.name: sm for sm in [ON, OFF, DAY, NIGHT]}

    @classmethod
    def get(cls, name: str) -> SettingMode:
        """Get :class:`SettingMode` by name.

        Args:
            name (str): Name of the setting mode

        Returns:
            The corresponding setting mode or None
        """
        return cls._VALUES[name]


@attr.s(frozen=True)
class QuickMode(Mode):
    """This class is a helper to check what is impacted by a quick mode.

    Quick modes are mode you can quickly set with the mobile app, most of the
    time, they have system wise impact. This is different from
    :class:`QuickVeto`, which will only impact one
    :class:`~pymultimatic.model.component.Component`.

    Args:
        name (str): Name of the quick mode.
        for_zone (bool): quick mode is applicable for a
            :class:`~pymultimatic.model.component.Zone`.
        for_room (bool): quick mode is applicable for a
            :class:`~pymultimatic.model.component.Room`.
        for_dhw (bool): quick mode is applicable for a
            :class:`~pymultimatic.model.component.HotWater` and
            :class:`~pymultimatic.model.component.Circulation`.

    Note:
        There is a default duration for quick modes (which can't be changed)
        but the API always returns `0` as remaining duration, so there is no
        way to know when quick mode will end.

    Vaillant documentation link about quick modes:
    http://elearning.vaillant.com/vrc700/ci/en/documents/uk/infopool/Operating_instructions.pdf
    """

    for_zone = attr.ib(type=bool)
    for_room = attr.ib(type=bool)
    for_dhw = attr.ib(type=bool)


class QuickModes:
    """Groups all :class:`QuickMode` together."""

    HOTWATER_BOOST = QuickMode('QM_HOTWATER_BOOST', False, False, True)
    """The advanced function heats the water in the domestic hot water cylinder
    once until the desired DHW circuit temperature set is reached or until you
    cancel the advanced function early.
    The heating installation will then return to the pre-set mode
    """

    VENTILATION_BOOST = QuickMode('QM_VENTILATION_BOOST', True, False, False)
    """This advanced function switches the zone off for 30 minutes.
    The frost protection function is activated, and hot water generation and
    circulation remain active.
    Ventilation is activated and works at the highest ventilation level.
    The advanced function is automatically deactivated after 30 minutes or
    if you cancel the advanced function early.
    The heating installation will then return to the pre-set mode.
    """

    ONE_DAY_AWAY = QuickMode('QM_ONE_DAY_AWAY', True, False, True)
    """Hot water generation and circulation are switched off and the frost
    protection is activated.
    The advanced function is automatically deactivated after 24:00 hours or if
    you cancel the advanced function first.
    The heating installation will then return to the pre-set mode.
    Ventilation is activated and works at the lowest ventilation level.
    """

    SYSTEM_OFF = QuickMode('QM_SYSTEM_OFF', True, True, True)
    """The heating function, hot water circuit and cooling are switched off.
    The frost protection function is activated.
    The circulation is switched off.
    Ventilation is activated and works at the lowest ventilation level.
    """

    ONE_DAY_AT_HOME = QuickMode('QM_ONE_DAY_AT_HOME', True, False, False)
    """This advanced function activates Automatic mode for one day with the
    settings for Sunday, as set using the Time programmes function.
    The advanced function is automatically deactivated after 24:00 hours or if
    you cancel the advanced function first.
    The heating installation will then return to the pre-set mode.
    The operating modes and time programme for
    :class:`~pymultimatic.model.component.Room` are not affected.
    """

    PARTY = QuickMode('QM_PARTY', True, False, False)
    """The advanced function brings the room temperature to the set desired
    Day temperature, in accordance with the set time periods.
    The advanced function is deactivated after six hours or if you cancel it
    before the six hours is up.
    The heating installation will then return to the pre-set mode.
    """

    HOLIDAY = QuickMode('QM_HOLIDAY', True, True, True)
    """
    """

    QUICK_VETO = QuickMode('QM_QUICK_VETO', False, False, False)
    """
    This advanced function activates a quick veto for one specific zone
    """

    _VALUES = {qm.name: qm for qm in [HOTWATER_BOOST, VENTILATION_BOOST,
                                      ONE_DAY_AWAY, SYSTEM_OFF,
                                      ONE_DAY_AT_HOME, PARTY, HOLIDAY,
                                      QUICK_VETO]}

    @classmethod
    def for_zone(cls) -> List[QuickMode]:
        """Get the list of :class:`QuickMode` applicable to
        :class:`~pymultimatic.model.component.Zone`.

        Returns:
            A list of quick mode applicable for
            :class:`~pymultimatic.model.component.Zone`.
        """
        sub_list = []

        for quick_mode in cls._VALUES.values():
            if quick_mode.for_zone:
                sub_list.append(quick_mode)

        return sub_list

    @classmethod
    def for_room(cls) -> List[QuickMode]:
        """Get the list of :class:`QuickMode` applicable to
        :class:`~pymultimatic.model.component.Room`.

        Returns:
            A list of quick mode applicable for
            :class:`~pymultimatic.model.component.Room`.
        """
        sub_list = []

        for quick_mode in cls._VALUES.values():
            if quick_mode.for_room:
                sub_list.append(quick_mode)

        return sub_list

    @classmethod
    def for_dhw(cls) -> List[QuickMode]:
        """Get the list of :class:`QuickMode` applicable to
        :class:`~pymultimatic.model.component.HotWater`
        and :class:`~pymultimatic.model.component.Circulation`.

        Returns:
            A list of quick mode applicable for
            :class:`~pymultimatic.model.component.HotWater` and
            :class:`~pymultimatic.model.component.Circulation`.
        """
        sub_list = []

        for quick_mode in cls._VALUES.values():
            if quick_mode.for_dhw:
                sub_list.append(quick_mode)

        return sub_list

    @classmethod
    def get(cls, name: str) -> QuickMode:
        """Get :class:`QuickMode` by name.

        Args:
            name (str): Name of the operating mode

        Returns:
            The corresponding operating mode or None
        """
        return cls._VALUES[name]


@attr.s
class ActiveMode:
    """Active mode will let you know the real target temperature and the real
    :class:`Mode` applied to a
    :class:`~pymultimatic.model.component.Component`.

    Note:
        This is based on the interpretation of the data coming from the API.

    It can be quite complex to know what is going on from a
    :class:`~pymultimatic.model.component.Component`,
    the API returns data at different places like for example:

    * the :class:`OperatingMode` (and the :class:`QuickVeto`) for a
      :class:`~pymultimatic.model.component.Component` are found at one place

    * the system wise :class:`QuickMode` is found somewhere else

    * not all the :class:`QuickMode` are applicable to all
      :class:`~pymultimatic.model.component.Component`

    So, if a :class:`~pymultimatic.model.component.Zone` is configured with
    :class:`OperatingModes.AUTO` mode and there is a
    :class:`QuickModes.QM_SYSTEM_OFF` active, the API won't tell you what is
    the real mode applied to that :class:`~pymultimatic.model.component.Zone`
    . In this case, the :class:`QuickModes.QM_SYSTEM_OFF` will actually take
    precedence and it will have impact on the target temperature.

    Args:
        target_temperature (Optional[float]): Target temperature the
            :class:`~pymultimatic.model.component.Component` tries to reach.
            Please note, there will be no `target_temperature` for
            :class:`~pymultimatic.model.component.Circulation`.
        current_mode (Mode): This is the mode that will be applied by your
            installation.
        sub_mode: (Mode): If `mode` is :class:`OperatingModes.AUTO`,
            then `sub_mode` will be populated with the actual configured mode,
            coming from
            :class:`~pymultimatic.model.timeprogram.TimeProgramDaySetting.mode`
            , otherwise it will be `None`
    """

    target_temperature = attr.ib(type=Optional[float])
    current_mode = attr.ib(type=Mode)
    sub_mode = attr.ib(type=Optional[Mode], default=None)


@attr.s
class HolidayMode:
    """Represents system's holiday mode.

    Args:
        is_active (bool): Indicates whether the user has activated the holiday
            mode or not.
        start_date (date): Start date of the holiday mode.
        end_date (date): End date of the holiday mode.
        target_temperature (float): Target temperature when the
            holiday mode is applied.

    Note:
        `is_active` doesn't necessarily means the holiday mode `is_applied`:
        the user can activate the holiday mode through the mobile app but with
        `end_date` and `start_date` in the future.
    """

    is_active = attr.ib(type=bool)
    start_date = attr.ib(type=Optional[date], default=None)
    end_date = attr.ib(type=Optional[date], default=None)
    target_temperature = attr.ib(type=Optional[float], default=None)

    @property
    def active_mode(self) -> Optional[ActiveMode]:
        """Optional[ActiveMode]: Get the :class:`ActiveMode` from holiday mode,
        if `is_applied`."""
        if self.is_applied:
            return ActiveMode(self.target_temperature, QuickModes.HOLIDAY)
        return None

    @property
    def is_applied(self) -> bool:
        """bool: Check if holiday mode is active **and** if today are between
        end_date and start_date.

        Note:
            Use this to know if the holiday mode is active and applied to the
            system.
        """
        return self.is_active \
            and self.start_date is not None \
            and self.end_date is not None \
            and self.start_date <= date.today() <= self.end_date



"""Boiler info module"""
from typing import Optional

import attr

# pylint: disable=too-few-public-methods
@attr.s
class BoilerInfo:
    """Information about the boiler.

    Args:
        water_pressure (float): Water pressure in bar
        current_temperature (float): Current temperature of the boiler.
    """

    water_pressure = attr.ib(type=Optional[float])
    current_temperature = attr.ib(type=Optional[float])



"""Low level connector module."""
import logging
import os
from typing import Optional, Dict, Any

import attr
import requests
from requests import Response

from . import ApiError, urls, defaults
from ..util import fileutils

_LOGGER = logging.getLogger('Connector')

_JSON_CONTENT_TYPE_HEADER = {'content-type': 'application/json'}


@attr.s
class ApiConnector:
    """This is the low level smart.vaillant.com API connector.

    This is returning the raw JSON from responses or an
    :exc:`~pymultimatic.api.error.ApiError` if something goes wrong
    (basically, when response error code is 4xx or 5xx).

    On the first call, the connector will login automatically (if
    :func:`~login` was not called previously).

    On following calls, the connector will re-use the cookie received from the
    API. If the connector receives an HTTP 401, it will clear the cookie and
    try to login before raising an `~pymultimatic.api.error.ApiError`. This
    also means the connector is able to reconnect automatically when cookie is
    outdated.

    Please use :mod:`~pymultimatic.api.urls` in order to generate URL to be
    passed to the connector.

    Args:
        user (str): User to login with.
        password (str): Password associated with the user.
        smartphone_id (str): This is required by the API to login.
        file_path: (str): Path where cookie is/will be stored.
    """

    _user = attr.ib(type=str)
    _password = attr.ib(type=str, repr=False)
    _smartphone_id = attr.ib(type=str, default=defaults.SMARTPHONE_ID)
    _file_path = attr.ib(type=str, default=defaults.FILES_PATH)
    _serial_number = attr.ib(type=Optional[str], default=None, init=False)
    _session = attr.ib(type=requests.Session, default=None, init=False)

    def __attrs_post_init__(self) -> None:
        self._file_path = os.path.expanduser(self._file_path)
        self._serial_number = self._load_serial_number_from_file()
        self._session = self._create_or_load_session()

    def login(self, force_login: bool = False) -> bool:
        """Log in to the API.

        By default, the ``connector`` will try to use cookie located under
        :attr:`file_path`.

        Args:
            force_login (bool): If set to ``True``, the connector will clear
                the cookie (if any) and start a new authentication, otherwise
                it will re-use the existing cookie.

        Returns:
            bool: True/False if authentication succeeded or not.
        """
        try:
            return self._authentication(force_login)
        except ApiError:
            return False

    def logout(self) -> None:
        """Get logged out of the API.

        It first sends a ``logout`` request to the API (it means cookies are
        invalidated).

        Second, cookies will be cleared, regardless of the result of the
        ``logout`` request.

        The connector will have to request a new token and ask for cookies if
        a new request is done.


        Raises:
            ApiError: When something went wrong with the API call.
        """
        response = None
        try:
            response = self._session.request('POST', urls.logout())
        except Exception as exc:
            raise ApiError("Error during logout", response) from exc
        finally:
            self._clear_session()

    def query(self, url: str, method: str = 'GET',
              payload: Optional[Dict[str, Any]] = None) -> Optional[Any]:
        """Call the vaillant API url with the chosen method, please use
        :mod:`~pymultimatic.api.urls` in order to generate the URL to be
        passed to the connector.

        Args:
            url (str): URL to call.
            method (str): HTTP method.
            payload (Dict[str, Any]): Payload to send.

        Returns:
            JSON response.

        Raises:
            ApiError: When something went wrong with the API call.
        """

        return self._safe_call(method, url, payload)

    def get(self, url: str) -> Optional[Any]:
        """Create a GET call to a vaillant API, please use
        :mod:`~pymultimatic.api.urls` in order to generate the URL to be
        passed to the connector.

        Args:
            url (str): URL to call.

        Returns:
            JSON response.

        Raises:
            ApiError: When something went wrong with the API call
        """
        return self.query(url)

    def put(self, url: str, payload: Optional[Dict[str, Any]] = None) \
            -> Optional[Any]:
        """Create a PUT call to a vaillant API, please use
        :mod:`~pymultimatic.api.urls` in order to generate the URL to be
        passed to the connector.

        Args:
            url (str): URL to call.
            payload (Dict[str, Any]): Payload to send.

        Returns:
            JSON response.

        Raises:
            ApiError: When something went wrong with the API call
        """
        return self.query(url, 'PUT', payload)

    def post(self, url: str, payload: Optional[Dict[str, Any]] = None) \
            -> Optional[Any]:
        """Create a POST call to a vaillant API, please use
        :mod:`~pymultimatic.api.urls` in order to generate the URL to be
        passed to the connector.

        Args:
            url (str): URL to call.
            payload (Dict[str, Any]): Payload to send.

        Returns:
            JSON response.

        Raises:
            ApiError: When something went wrong with the API call
        """
        return self.query(url, 'POST', payload)

    def delete(self, url: str) -> Optional[Any]:
        """Create a DELETE call to a vaillant API, please use
        :mod:`~pymultimatic.api.urls` in order to generate the URL to be
        passed to the connector.

        Args:
            url (str): URL to call.

        Returns:
            JSON response.

        Raises:
            ApiError: When something went wrong with the API call
        """
        return self.query(url, 'DELETE')

    def _safe_call(self, method: str, url: str,
                   payload: Optional[Dict[str, Any]] = None,
                   re_login: bool = False) -> Optional[Any]:
        """Call the API using selected :attr:`method`, attr:`url` and
        :attr:`payload`.

        This is ``safe`` in a sense the connector ensure you are logged in.

        The replacement of ``serial_number`` placeholder in
        :mod:`~pymultimatic.api.urls` is done here.
        """
        response: Optional[Response] = None
        safe_url: Optional[str] = None

        try:
            self._authentication(re_login)

            safe_url = url.format(serial_number=self._serial_number)
            header = None if payload is None else _JSON_CONTENT_TYPE_HEADER
            response = self._session.request(method,
                                             safe_url,
                                             json=payload,
                                             headers=header)

            if response is not None:
                if response.status_code > 399:
                    if not re_login and response.status_code == 401:
                        _LOGGER.debug('Call to %s failed with HTTP 401, '
                                      'will try to re-login', safe_url)
                        return self._safe_call(method, url, payload, True)

                    raise ApiError(
                        'Received error from server url: {} and method {}'
                        .format(safe_url, method),
                        response,
                        payload)

                return self._return_json(response)

            raise ApiError('Cannot {} url: {}'.format(method, safe_url),
                           response,
                           payload)
        except ApiError:
            raise
        except Exception as exc:
            raise ApiError(
                'Cannot {} url: {}'
                .format(method, safe_url if safe_url else url),
                response,
                payload) from exc

    # pylint: disable=no-self-use
    def _return_json(self, response: Response) -> Optional[Any]:
        if response.text:
            return response.json()
        return None

    def _authentication(self, force_login: bool = False) -> bool:
        try:
            if force_login:
                self._clear_session()

            if not self._session.cookies:
                self._session = self._create_or_load_session()
                self._serial_number = self._load_serial_number_from_file()

                if not self._session.cookies:
                    _LOGGER.info('No previous session found, will try to log '
                                 'in with username: %s and smartphoneId: %s',
                                 self._user, self._smartphone_id)

                    auth_token = self._request_token()
                    self._get_cookies(auth_token)

            if not self._serial_number:
                self._get_serial_number()

            return True
        except ApiError:
            raise
        except Exception as exc:
            raise ApiError('Error during login', None) from exc

    def _request_token(self) -> str:
        params = {
            "smartphoneId": self._smartphone_id,
            "username": self._user,
            "password": self._password
        }

        try:
            response = self._session.post(urls.new_token(), json=params,
                                          headers=_JSON_CONTENT_TYPE_HEADER)
            if response.status_code == 200:
                _LOGGER.debug('Token generation successful')
                return str(response.json()['body']['authToken'])

            params['password'] = '*****'
            raise ApiError('Authentication failed', response, params)

        except ApiError:
            raise
        except Exception as exc:
            raise ApiError('Error during authentication', None) from exc

    def _get_cookies(self, auth_token: str) -> None:
        params = {
            "smartphoneId": self._smartphone_id,
            "username": self._user,
            "authToken": auth_token
        }

        try:
            response = self._session.post(urls.authenticate(), json=params,
                                          headers=_JSON_CONTENT_TYPE_HEADER)

            if response.status_code == 200:
                self._session.cookies = response.cookies
                _LOGGER.debug('Cookie successfully retrieved %s',
                              self._session.cookies)
                self._save_cookies_to_file()
            else:
                params["authToken"] = "******"
                raise ApiError('Cannot get cookies', response, params)
        except ApiError:
            raise
        except Exception as exc:
            raise ApiError('Error while getting cookies', None) from exc

    def _get_serial_number(self) -> None:
        try:
            response = self._session.get(urls.facilities_list())

            if response.status_code == 200:
                _LOGGER.debug('Serial number successfully retrieved')
                self._serial_number = \
                    response.json()['body']['facilitiesList'][0][
                        'serialNumber']
                self._save_serial_number_to_file()
            else:
                raise ApiError('Cannot get serial number', response)
        except ApiError:
            raise
        except Exception as exc:
            raise ApiError('Cannot get serial number', None) from exc

    def _create_or_load_session(self) -> requests.Session:
        session = requests.Session()
        cookies = self._load_cookies_from_file()
        _LOGGER.debug('Found cookies %s', cookies)
        if cookies is not None:
            session.cookies = cookies
        return session

    def _clear_session(self) -> None:
        self._clear_cookie()
        self._clear_serial_number()
        self._session.close()
        self._session = requests.session()
        fileutils.delete_dir(self._file_path)

    def _save_cookies_to_file(self) -> None:
        fileutils.save_to_file(self._session.cookies, self._file_path,
                               defaults.COOKIE_FILE_NAME)

    def _save_serial_number_to_file(self) -> None:
        fileutils.save_to_file(self._serial_number, self._file_path,
                               defaults.SERIAL_NUMBER_FILE_NAME)

    def _load_cookies_from_file(self) -> Any:
        return fileutils.load_from_file(self._file_path,
                                        defaults.COOKIE_FILE_NAME)

    def _load_serial_number_from_file(self) -> Optional[str]:
        result = fileutils.load_from_file(self._file_path,
                                          defaults.SERIAL_NUMBER_FILE_NAME)

        if result:
            return str(result)
        return None

    def _clear_cookie(self) -> None:
        fileutils.delete_file(self._file_path, defaults.COOKIE_FILE_NAME)

    def _clear_serial_number(self) -> None:
        fileutils.delete_file(self._file_path,
                              defaults.SERIAL_NUMBER_FILE_NAME)
        self._serial_number = None



"""Full system coming from vaillant API."""
import datetime
from typing import List, Optional, Dict

import attr

from . import ActiveMode, HolidayMode, HotWater, Room, Zone, BoilerStatus, \
    Circulation, QuickMode, QuickModes, Error, SystemStatus, BoilerInfo, \
    constants, SettingModes


# pylint: disable=too-many-instance-attributes
@attr.s
class System:
    """This class represents the main class to manipulate vaillant system. It
    groups all the information about the system.

    Args:
        holiday_mode (HolidayMode): Holiday mode.
        system_status (SystemStatus): Status of the system.
        boiler_status (BoilerStatus): Status of the boiler.
        zones (List[Zone]): List of zone.
        rooms (List[Room]): List of room.
        hot_water (HotWater): If hot water is present, it's available here.
        circulation (Circulation): If circulation is present, it's available
            here.
        outdoor_temperature (float): Outdoor temperature
        quick_mode (QuickMode):  If any quick mode is running, it's available
            here.
        errors (List[Error]): If there are errors, you can find them here.
        boiler_info (BoilerInfo): Information (pressure, temperature) about
            the boiler
    """

    holiday_mode = attr.ib(type=HolidayMode)
    system_status = attr.ib(type=SystemStatus)
    boiler_status = attr.ib(type=Optional[BoilerStatus])
    zones = attr.ib(type=List[Zone])
    rooms = attr.ib(type=List[Room])
    hot_water = attr.ib(type=Optional[HotWater])
    circulation = attr.ib(type=Optional[Circulation])
    outdoor_temperature = attr.ib(type=Optional[float])
    quick_mode = attr.ib(type=Optional[QuickMode])
    errors = attr.ib(type=List[Error])
    boiler_info = attr.ib(type=Optional[BoilerInfo])
    _zones = attr.ib(type=Dict[str, Zone], default=dict(), init=False)
    _rooms = attr.ib(type=Dict[str, Room], default=dict(), init=False)

    def __attrs_post_init__(self) -> None:
        """Post init from attrs."""
        if self.holiday_mode is None:
            self.holiday_mode = HolidayMode(False)

        if self.zones:
            self._zones = dict((zone.id, zone) for zone in self.zones)

        if self.rooms:
            self._rooms = dict((room.id, room) for room in self.rooms)

    def set_zone(self, zone_id: str, zone: Zone) -> None:
        """Set :class:`~pymultimatic.model.component.Zone` for the given id.

        Args:
            zone_id (str): id of the zone
            zone (Zone): the zone to set

        Returns:
            None
        """
        self._zones[zone_id] = zone
        self.zones = list(self._zones.values())

    def set_room(self, room_id: str, room: Room) -> None:
        """Set :class:`~pymultimatic.model.component.Room` for the given id.

        Args:
            room_id (str): id of the room
            room (Room): the room to set

        Returns:
            None
        """
        self._rooms[room_id] = room
        self.rooms = list(self._rooms.values())

    def get_active_mode_zone(self, zone: Zone) -> ActiveMode:
        """Get the current
        :class:`~pymultimatic.model.mode.ActiveMode` for a
        :class:`~pymultimatic.model.component.Zone`. This is the only way to
        get the real one.

        Args:
            zone (Zone): The zone you want the active mode for.

        Returns:
            ActiveMode: The active mode.
        """
        mode: ActiveMode = zone.active_mode

        # Holiday mode takes precedence over everything
        if self.holiday_mode.active_mode:
            mode = self.holiday_mode.active_mode

        # Global system quick mode takes over zone settings
        if self.quick_mode and self.quick_mode.for_zone:
            if self.quick_mode == QuickModes.VENTILATION_BOOST:
                mode = ActiveMode(Zone.MIN_TARGET_TEMP, self.quick_mode)

            if self.quick_mode == QuickModes.ONE_DAY_AWAY:
                mode = ActiveMode(Zone.MIN_TARGET_TEMP, self.quick_mode)

            if self.quick_mode == QuickModes.SYSTEM_OFF:
                mode = ActiveMode(Zone.MIN_TARGET_TEMP, self.quick_mode)

            if self.quick_mode == QuickModes.ONE_DAY_AT_HOME:
                today = datetime.datetime.now()
                sunday = today - datetime.timedelta(days=today.weekday() - 6)

                time_program = zone.time_program.get_for(sunday)
                target_temp = zone.target_temperature
                if time_program.setting == SettingModes.NIGHT:
                    target_temp = zone.target_min_temperature

                mode = ActiveMode(target_temp, self.quick_mode)

            if self.quick_mode == QuickModes.PARTY:
                mode = ActiveMode(zone.target_temperature, self.quick_mode)

        return mode

    def get_active_mode_room(self, room: Room) -> ActiveMode:
        """Get the current
        :class:`~pymultimatic.model.mode.ActiveMode` for a
        :class:`~pymultimatic.model.component.Room`. This is the only way to
        get the real one.

        Args:
            room (Room): The room you want the active mode for.

        Returns:
            ActiveMode: The active mode.
        """
        # Holiday mode takes precedence over everything
        if self.holiday_mode.active_mode:
            return self.holiday_mode.active_mode

        # Global system quick mode takes over room settings
        if self.quick_mode and self.quick_mode.for_room:
            if self.quick_mode == QuickModes.SYSTEM_OFF:
                return ActiveMode(Room.MIN_TARGET_TEMP, self.quick_mode)

        return room.active_mode

    def get_active_mode_circulation(self,
                                    circulation: Optional[Circulation] = None)\
            -> Optional[ActiveMode]:
        """Get the current
        :class:`~pymultimatic.model.mode.ActiveMode` for a
        :class:`~pymultimatic.model.component.Circulation`. This is the only
        way to get the real one.

        Args:
            circulation (Circulation): The circulation you want the active mode
            for. If ``None`` is passed, it will pick up the circulation from
            the system.

        Returns:
            ActiveMode: The active mode.
        """
        if not circulation:
            circulation = self.circulation

        if circulation:
            if self.holiday_mode.active_mode:
                active_mode = self.holiday_mode.active_mode
                active_mode.target_temperature = None
                return active_mode

            if self.quick_mode and self.quick_mode.for_dhw:
                return ActiveMode(None, self.quick_mode)

            return circulation.active_mode
        return None

    def get_active_mode_hot_water(self, hot_water: Optional[HotWater] = None)\
            -> Optional[ActiveMode]:
        """Get the current
        :class:`~pymultimatic.model.mode.ActiveMode` for a
        :class:`~pymultimatic.model.component.HotWater`. This is the only way
        to get the real one.

        Args:
            hot_water (HotWater): The hot water you want the active mode
            for. If ``None`` is passed, it will pick up the hot water from the
            system.

        Returns:
            ActiveMode: The active mode.
        """
        if not hot_water:
            hot_water = self.hot_water

        if hot_water:
            if self.holiday_mode.active_mode:
                active_mode = self.holiday_mode.active_mode
                active_mode.target_temperature = \
                    constants.FROST_PROTECTION_TEMP
                return active_mode

            if self.quick_mode and self.quick_mode.for_dhw:
                if self.quick_mode == QuickModes.HOTWATER_BOOST:
                    return ActiveMode(hot_water.target_temperature,
                                      self.quick_mode)

                if self.quick_mode == QuickModes.SYSTEM_OFF:
                    return ActiveMode(constants.FROST_PROTECTION_TEMP,
                                      self.quick_mode)

                if self.quick_mode == QuickModes.ONE_DAY_AWAY:
                    return ActiveMode(constants.FROST_PROTECTION_TEMP,
                                      self.quick_mode)

            return hot_water.active_mode
        return None



"""Mapped model from the API."""
# pylint: disable=cyclic-import
from .mode import Mode, OperatingMode, OperatingModes, QuickMode, QuickModes, \
    QuickVeto, HolidayMode, ActiveMode, SettingMode, SettingModes
from .timeprogram import TimeProgram, TimeProgramDay, TimePeriodSetting
from .component import Component, Circulation, HotWater, Device, Room, Zone
from .status import Error, BoilerStatus, SystemStatus
from .boilerinfo import BoilerInfo
from .system import System
from .syncstate import SyncState



"""Mappers from json to model classes."""
from datetime import datetime
from typing import Optional, List, Any

from . import BoilerStatus, Circulation, Device, HolidayMode, HotWater, \
    QuickMode, QuickModes, QuickVeto, Room, TimeProgram, TimeProgramDay, \
    TimePeriodSetting, Zone, OperatingModes, Error, SystemStatus, SyncState, \
    SettingModes, BoilerInfo


_DATE_FORMAT = "%Y-%m-%d"


def map_quick_mode(full_system) -> Optional[QuickMode]:
    """Map *quick mode*."""
    if full_system:
        quick_mode = full_system.get("body", dict())\
            .get("configuration", dict()).get("quickmode")
        if quick_mode:
            mode = QuickModes.get(quick_mode.get("quickmode"))
            if mode != QuickModes.QUICK_VETO:
                return mode
    return None


def map_outdoor_temp(full_system) -> Optional[float]:
    """get *outdoor_temperature*."""
    if full_system:
        raw_temp = full_system.get("body", dict()).get("status", dict())\
            .get('outside_temperature')
        if raw_temp is not None:
            return float(raw_temp)
    return None


def map_rooms(raw_rooms) -> List[Room]:
    """Map *rooms*."""
    rooms: List[Room] = []
    if raw_rooms:
        for raw_room in raw_rooms.get("body", dict()).get("rooms", list()):
            room = map_room(raw_room)
            if room:
                rooms.append(room)

    return rooms


def map_room(raw_room) -> Optional[Room]:
    """Map *room*."""
    if raw_room:
        raw_room = raw_room.get("body", raw_room)

        if raw_room:
            config = raw_room.get("configuration", dict())

            component_id = raw_room.get("roomIndex")
            child_lock = config.get("childLock")
            target_temp = config.get("temperatureSetpoint")
            current_temp = config.get("currentTemperature")
            devices = map_devices(config.get("devices"))
            window_open = config.get("isWindowOpen")
            name = config.get("name")
            operation_mode = OperatingModes.get(config.get("operationMode"))
            humidity = config.get('currentHumidity')

            raw_quick_veto = config.get("quickVeto")
            quick_veto = None
            if raw_quick_veto:
                quick_veto = QuickVeto(
                    raw_quick_veto.get("remainingDuration"),
                    config.get("temperatureSetpoint"))

            time_program = map_time_program(raw_room.get("timeprogram"))

            return Room(component_id, name, time_program, current_temp,
                        target_temp, operation_mode, quick_veto,
                        child_lock, window_open, devices, humidity)
    return None


def map_devices(raw_devices) -> List[Device]:
    """Map *devices* of a room."""
    devices = []
    if raw_devices:
        for raw_device in raw_devices:
            name = raw_device.get("name")
            device_type = raw_device.get("deviceType")
            battery_low = raw_device.get("isBatteryLow")
            radio_out_of_reach = raw_device.get("isRadioOutOfReach")
            sgtin = raw_device.get("sgtin")
            devices.append(
                Device(name, sgtin, device_type, battery_low,
                       radio_out_of_reach))

    return devices


def map_time_program(raw_time_program, key: Optional[str] = None)\
        -> TimeProgram:
    """Map *time program*."""
    result = {}
    if raw_time_program:
        result["monday"] = map_time_program_day(
            raw_time_program.get("monday"), key)
        result["tuesday"] = map_time_program_day(
            raw_time_program.get("tuesday"), key)
        result["wednesday"] = map_time_program_day(
            raw_time_program.get("wednesday"), key)
        result["thursday"] = map_time_program_day(
            raw_time_program.get("thursday"), key)
        result["friday"] = map_time_program_day(
            raw_time_program.get("friday"), key)
        result["saturday"] = map_time_program_day(
            raw_time_program.get("saturday"), key)
        result["sunday"] = map_time_program_day(
            raw_time_program.get("sunday"), key)

    return TimeProgram(result)


def map_time_program_day(raw_time_program_day, key: Optional[str] = None) \
        -> TimeProgramDay:
    """Map *time program day* and *time program day settings*."""
    settings = []
    if raw_time_program_day:
        for time_setting in raw_time_program_day:
            start_time = time_setting.get("startTime")
            target_temp = time_setting.get("temperatureSetpoint")

            mode = None
            if key:
                mode = SettingModes.get(time_setting.get(key))

            settings.append(
                TimePeriodSetting(start_time, target_temp, mode))

    return TimeProgramDay(settings)


def map_holiday_mode(full_system) -> HolidayMode:
    """Map *holiday mode*."""
    mode = HolidayMode(False)
    if full_system:
        raw_holiday_mode = full_system.get("body", dict())\
            .get("configuration", dict()).get("holidaymode")

        if raw_holiday_mode:
            mode.is_active = bool(raw_holiday_mode.get("active"))
            mode.target_temperature = float(raw_holiday_mode
                                            .get("temperature_setpoint"))
            mode.start_date = datetime.strptime(
                raw_holiday_mode.get("start_date"), _DATE_FORMAT).date()
            mode.end_date = datetime.strptime(
                raw_holiday_mode.get("end_date"), _DATE_FORMAT).date()

    return mode


def map_boiler_status(hvac_state) -> Optional[BoilerStatus]:
    """Map *boiler status."""
    if hvac_state:
        hvac_state_info = _find_hvac_message_status(hvac_state)
        if hvac_state_info:
            last_update = _datetime_mandatory(hvac_state_info.get("timestamp"))
            device_name = str(hvac_state_info.get("deviceName"))
            code = str(hvac_state_info.get("statusCode"))
            title = str(hvac_state_info.get("title"))
            description = str(hvac_state_info.get("description"))
            hint = str(hvac_state_info.get("hint"))
            return BoilerStatus(device_name, title, code, description,
                                last_update, hint)

    return None


def map_boiler_info(live_report) -> Optional[BoilerInfo]:
    """Map boiler info"""
    if live_report:
        water_pressure = _find_water_pressure_report(live_report)
        boiler_temperature = _find_boiler_temperature_report(live_report)
        return BoilerInfo(water_pressure, boiler_temperature)
    return None


def map_system_status(hvac_state) -> SystemStatus:
    """Map *system status*."""
    meta = hvac_state.get('meta', dict())
    online = meta.get('onlineStatus', dict()).get('status')
    update = meta.get('firmwareUpdateStatus', dict()).get('status')
    return SystemStatus(online, update)


def map_zones(full_system) -> List[Zone]:
    """Map *zones*."""
    zones = []
    if full_system:
        for raw_zone in full_system.get("body", dict()).get("zones", list()):
            zone = map_zone(raw_zone)
            if zone:
                zones.append(zone)

    return zones


def map_zone(raw_zone) -> Optional[Zone]:
    """Map *zones*."""
    raw_zone_body = raw_zone.get("body")

    if raw_zone_body:
        raw_zone = raw_zone_body

    if raw_zone:
        heating = raw_zone.get("heating", dict())
        configuration = raw_zone.get("configuration", dict())
        heating_configuration = heating.get("configuration", dict())

        zone_id = raw_zone.get("_id")
        operation_mode = OperatingModes.get(heating_configuration.get("mode"))
        target_temp = heating_configuration.get("setpoint_temperature")
        target_min_temp = heating_configuration.get("setback_temperature")
        time_program = map_time_program(heating.get("timeprogram"), "setting")

        name = configuration.get("name", "").strip()
        current_temperature = configuration.get("inside_temperature")
        active_function = configuration.get("active_function")

        quick_veto = _map_quick_veto_zone(configuration.get("quick_veto"))

        rbr = raw_zone.get("currently_controlled_by", dict())\
                      .get("name", "") == "RBR"

        return Zone(zone_id, name, time_program, current_temperature,
                    target_temp, operation_mode, quick_veto,
                    target_min_temp, active_function, rbr)

    return None


def map_hot_water(full_system, live_report) -> Optional[HotWater]:
    """Map *hot water*."""
    hot_water_list = None
    if full_system:
        hot_water_list = full_system.get("body", dict()).get("dhw")

    if hot_water_list:
        raw_hot_water = hot_water_list[0].get("hotwater")
        dwh_id = hot_water_list[0].get("_id")

        if raw_hot_water:
            return _map_hot_water(raw_hot_water, dwh_id, live_report)

    return None


def map_hot_water_alone(raw_hot_water, dhw_id: str, live_report) \
        -> Optional[HotWater]:
    """Map *hot water*."""
    if raw_hot_water:
        raw_hot_water_body = raw_hot_water.get("body", dict())
        return _map_hot_water(raw_hot_water_body, dhw_id, live_report)
    return None


def map_circulation(full_system) -> Optional[Circulation]:
    """Map *circulation*."""
    if full_system:
        hot_water_list = full_system.get("body", dict()).get("dhw", list())

        if hot_water_list:
            raw_circulation = hot_water_list[0].get("circulation")
            dhw_id = hot_water_list[0].get("_id")

            if raw_circulation:
                return _map_circulation(raw_circulation, dhw_id)
    return None


def map_circulation_alone(raw_circulation, dhw_id: str)\
        -> Optional[Circulation]:
    """Map *circulation*."""
    if raw_circulation:
        raw_circulation_body = raw_circulation.get("body", dict())
        return _map_circulation(raw_circulation_body, dhw_id)
    return None


def map_errors(hvac_state) -> List[Error]:
    """Map *errors*."""
    errors = []
    for error in hvac_state.get("body", dict()).get("errorMessages",
                                                    list()):
        if error.get("type") == "ERROR":
            errors.append(Error(error.get('deviceName'),
                                error.get('title'),
                                error.get('statusCode'),
                                error.get('description'),
                                _datetime_mandatory(error.get('timestamp'))))
    return errors


def map_hvac_sync_state(hvac_state) -> Optional[SyncState]:
    """Map """
    if hvac_state:
        states = hvac_state.get('meta', dict()).get('syncState', list())
        if states:
            return _map_state(states[0])
    return None


def _map_state(raw_state) -> Optional[SyncState]:
    if raw_state:
        state = str(raw_state.get('state'))
        timestamp = _datetime_mandatory(raw_state.get('timestamp'))
        link = raw_state.get('link', dict()).get('resourceLink')
        return SyncState(state, timestamp, link)
    return None


def _map_hot_water(raw_hot_water, dhw_id: str, live_report) \
        -> Optional[HotWater]:
    if raw_hot_water:
        target_temp = raw_hot_water.get("configuration", dict())\
            .get("temperature_setpoint")

        raw_operation_mode = raw_hot_water.get("configuration", dict())\
            .get("operation_mode")

        operation_mode = OperatingModes.get(raw_operation_mode)

        time_program = map_time_program(raw_hot_water.get("timeprogram",
                                                          dict()), "mode")

        current_temp = None
        name = None

        if live_report:
            dhw_report = _find_dhw_temperature_report(live_report)

            if dhw_report:
                current_temp = dhw_report.get("value")
                name = dhw_report.get("name")

        return HotWater(dhw_id, name, time_program, current_temp, target_temp,
                        operation_mode)
    return None


def _map_circulation(raw_circulation, circulation_id: str) -> Circulation:
    name = "Circulation"
    time_program = map_time_program(raw_circulation.get("timeprogram"),
                                    "setting")
    raw_operation_mode = raw_circulation.get("configuration", dict())\
        .get("operationMode")

    operation_mode = OperatingModes.get(raw_operation_mode)

    return Circulation(circulation_id, name, time_program, operation_mode)


def _find_hvac_message_status(hvac_state) -> Optional[Any]:
    for message in hvac_state.get("body", dict()).get("errorMessages", list()):
        if message.get("type") == "STATUS":
            return message
    return None


def _find_water_pressure_report(live_report) -> Optional[float]:
    if live_report:
        for device in live_report.get("body", dict()).get("devices", list()):
            for report in device.get("reports", list()):
                if report.get("associated_device_function") == "HEATING" \
                        and report.get("_id") == "WaterPressureSensor":
                    return float(report.get("value"))
    return None


def _find_boiler_temperature_report(live_report) -> Optional[float]:
    if live_report:
        for device in live_report.get("body", dict()).get("devices", list()):
            for report in device.get("reports", list()):
                if report.get("associated_device_function") == "HEATING" \
                        and (report.get("_id") == "FlowTemperatureSensor"
                             or report.get("_id") == "FlowTemperatureVF1"):
                    return float(report.get("value"))
    return None


def _find_dhw_temperature_report(live_report) -> Optional[Any]:
    if live_report:
        for device in live_report.get("body", dict()).get("devices", list()):
            for report in device.get("reports", list()):
                if report.get("associated_device_function") == "DHW" \
                        and report.get("_id") == \
                        "DomesticHotWaterTankTemperature":
                    return report
    return None


def _map_quick_veto_zone(raw_quick_veto) -> Optional[QuickVeto]:
    if raw_quick_veto and raw_quick_veto.get("active"):
        # No way to find start_date, Quick veto on zone lasts 6 hours
        return QuickVeto(None, raw_quick_veto.get("setpoint_temperature"))
    return None


def _datetime(timestamp: Optional[int]) -> Optional[datetime]:
    if timestamp:
        return datetime.fromtimestamp(timestamp / 1000)
    return None


def _datetime_mandatory(timestamp: int) -> datetime:
    return datetime.fromtimestamp(timestamp / 1000)



"""Convenient manager to easily gets data from API."""
import logging
from datetime import date, timedelta
from typing import Optional, List

from .api import ApiConnector, urls, payloads, defaults, ApiError
from .model import mapper, System, HotWater, QuickMode, QuickVeto, Room, \
    Zone, OperatingMode, Circulation, OperatingModes, constants

_LOGGER = logging.getLogger('SystemManager')


# pylint: disable=too-many-public-methods
class SystemManager:
    """This is a convenient manager to help interact with vaillant API.

    The manager is calling the
    :class:`~pymultimatic.api.connector.ApiConnector`, the manager will
    provide to correct :mod:`~pymultimatic.api.payloads` and
    :mod:`~pymultimatic.api.urls` according to what you want to do.

    The manager is throwing :class:`~pymultimatic.api.error.ApiError`
    without altering it.

    Args:
        user (str): User to login with.
        password (str): Password associated with the user.
        smartphone_id (str): This is required by the API to login.
        file_path: (str): Path where cookie is/will be stored.
    """

    def __init__(self, user: str, password: str,
                 smartphone_id: str = defaults.SMARTPHONE_ID,
                 file_path: str = defaults.FILES_PATH):
        self._connector: ApiConnector = \
            ApiConnector(user, password, smartphone_id, file_path)

    def login(self, force_login: bool = False) -> bool:
        """Try to login to the API, see
        :func:`~pymultimatic.api.connector.ApiConnector.login`

        Args:
            force_login (bool): Whether the login should be forced or not.

        Returns:
            bool: True/False if authentication succeeded or not.
        """
        return self._connector.login(force_login)

    # pylint: disable=too-many-locals
    def get_system(self) -> System:
        """Get the full :class:`~pymultimatic.model.system.System`. It may
        take some times, it actually does 3 or 4 API calls, depending on your
        system configuration.

        Returns:
            System: the full system.
        """
        full_system = self._connector.get(urls.system())
        live_report = self._connector.get(urls.live_report())
        hvac_state = self._connector.get(urls.hvac())

        holiday_mode = mapper.map_holiday_mode(full_system)
        boiler_status = mapper.map_boiler_status(hvac_state)
        system_status = mapper.map_system_status(hvac_state)

        zones = mapper.map_zones(full_system)

        rooms: List[Room] = []
        for zone in zones:
            if zone.rbr:
                rooms = mapper.map_rooms(self._connector.get(urls.rooms()))
                break

        hot_water = mapper.map_hot_water(full_system, live_report)
        circulation = mapper.map_circulation(full_system)

        outdoor_temp = mapper.map_outdoor_temp(full_system)
        quick_mode = mapper.map_quick_mode(full_system)
        errors = mapper.map_errors(hvac_state)
        boiler_info = mapper.map_boiler_info(live_report)

        return System(holiday_mode, system_status, boiler_status, zones, rooms,
                      hot_water, circulation, outdoor_temp, quick_mode, errors,
                      boiler_info)

    def get_hot_water(self, dhw_id: str) -> Optional[HotWater]:
        """Get the :class:`~pymultimatic.model.component.HotWater`
        information for the given id.

        Args:
            dhw_id (str): domestic hot water id (same as
                :func:`get_circulation`).

        Returns:
            HotWater: the hot water information, if any.
        """

        full_system = self._connector.get(urls.hot_water(dhw_id))
        live_report = self._connector.get(urls.live_report())
        return mapper.map_hot_water_alone(full_system, dhw_id, live_report)

    def get_room(self, room_id: str) -> Optional[Room]:
        """Get the :class:`~pymultimatic.model.component.Room` information
        for the given id.

        Args:
            room_id (str): Id of the room, this is actually an index, 1,2,3,4
            ... depending of the number of rooms you have.

        Returns:
            Room: the room information, if any.
        """
        new_room = self._connector.get(urls.room(room_id))
        return mapper.map_room(new_room)

    def get_zone(self, zone_id: str) -> Optional[Zone]:
        """"Get the :class:`~pymultimatic.model.component.Zone` information
        for the given id.

        Args:
            zone_id (str): Name of the room, basically, you set it through the
                VRC 700.

        Returns:
            Zone: the zone information, if any.
        """
        new_zone = self._connector.get(urls.zone(zone_id))
        return mapper.map_zone(new_zone)

    def get_circulation(self, dhw_id: str) \
            -> Optional[Circulation]:
        """"Get the :class:`~pymultimatic.model.component.Circulation`
        information for the given id.

        Args:
            dhw_id (str): domestic hot water id (same as :func:`get_hot_water`)

        Returns:
            Circulation: the circulation information, if any.
        """
        new_circulation = self._connector.get(urls.circulation(dhw_id))
        return mapper.map_circulation_alone(new_circulation, dhw_id)

    def set_hot_water_setpoint_temperature(self, dhw_id: str,
                                           temperature: float) -> None:
        """This set the target temperature for *hot water*."""
        _LOGGER.debug("Will set dhw target temperature to %s",
                      temperature)
        self._connector.put(
            urls.hot_water_temperature_setpoint(dhw_id),
            payloads.hotwater_temperature_setpoint(self._round(temperature)))

    def set_hot_water_operating_mode(self, dhw_id: str,
                                     new_mode: OperatingMode) -> None:
        """Set new operating mode for
        :class:`~pymultimatic.model.component.HotWater`. The mode should be
        listed here :class:`~pymultimatic.model.component.HotWater.MODES`
        otherwise it won't have any effect.


        Note:
            To set :class:`~pymultimatic.model.mode.QuickMode`, you
            have to use :func:`set_quick_mode`.

        Note:
            This call won't have any effect if there is a
            :class:`~pymultimatic.model.mode.QuickMode` activated, if you
            want to remove the quick mode, use :func:`remove_quick_mode`.

        Args:
            dhw_id (str): domestic hot water id.
            new_mode (OperatingMode): The new mode to set.
        """
        _LOGGER.debug("Will try to set hot water mode to %s", new_mode)

        if new_mode in HotWater.MODES:
            _LOGGER.debug("New mode is %s", new_mode)
            self._connector.put(
                urls.hot_water_operating_mode(dhw_id),
                payloads.hot_water_operating_mode(new_mode.name))
        else:
            _LOGGER.debug("New mode is not available for hot water %s",
                          new_mode)

    def set_room_operating_mode(self, room_id: str, new_mode: OperatingMode) \
            -> None:
        """Set new operating mode for
        :class:`~pymultimatic.model.component.Room`. The mode should be
        listed here :class:`~pymultimatic.model.component.Room.MODES`
        otherwise it won't have any effect.


        Note:
            To set :class:`~pymultimatic.model.mode.QuickMode`, you
            have to use :func:`set_quick_mode`.

        Note:
            This call won't have any effect if there is a
            :class:`~pymultimatic.model.mode.QuickMode` activated, if you
            want to remove the quick mode, use :func:`remove_quick_mode`.

        Note:
            To set :class:`~pymultimatic.model.mode.QuickVeto`, you have to
            use :func:`set_room_quick_veto`

        Args:
            room_id (str): id of the room.
            new_mode (OperatingMode): The new mode to set.
        """
        if new_mode in Room.MODES and new_mode != OperatingModes.QUICK_VETO:
            _LOGGER.debug("New mode is %s", new_mode)
            self._connector.put(urls.room_operating_mode(room_id),
                                payloads.room_operating_mode(
                                    new_mode.name))
        else:
            _LOGGER.debug("mode is not available for room %s", new_mode)

    def set_zone_operating_mode(self, zone_id: str, new_mode: OperatingMode) \
            -> None:
        """Set new operating mode for
        :class:`~pymultimatic.model.component.Zone`. The mode should be
        listed here :class:`~pymultimatic.model.component.Zone.MODES`
        otherwise it won't have any effect.


        Note:
            To set :class:`~pymultimatic.model.mode.QuickMode`, you
            have to use :func:`set_quick_mode`.

        Note:
            This call won't have any effect if there is a
            :class:`~pymultimatic.model.mode.QuickMode` activated, if you
            want to remove the quick mode, use :func:`remove_quick_mode`.

        Note:
            To set :class:`~pymultimatic.model.mode.QuickVeto`, you have to
            use :func:`set_zone_quick_veto`

        Args:
            zone_id (str): id of the zone.
            new_mode (OperatingMode): The new mode to set.
        """
        if new_mode in Zone.MODES and new_mode != OperatingModes.QUICK_VETO:
            _LOGGER.debug("New mode is %s", new_mode)
            self._connector.put(urls.zone_heating_mode(zone_id),
                                payloads.zone_operating_mode(new_mode.name))
        else:
            _LOGGER.debug("mode is not available for zone %s", new_mode)

    def set_quick_mode(self, quick_mode: QuickMode) -> None:
        """Set a :class:`~pymultimatic.model.mode.QuickMode` system wise.

        This will override the current
        :class:`~pymultimatic.model.mode.QuickMode`, if any.

        Args:
            quick_mode (QuickMode): the quick mode to set, see
                :class:`~pymultimatic.model.mode.QuickModes`
        """
        self._connector.put(urls.system_quickmode(),
                            payloads.quickmode(quick_mode.name))

    def remove_quick_mode(self) -> None:
        """Removes current :class:`~pymultimatic.model.mode.QuickMode`.

        Note:
            if there is not :class:`~pymultimatic.model.mode.QuickMode` set,
            the API returns an error (HTTP 409). **This error is swallowed by
            the manager**, so you don't have to handle it."""
        try:
            self._connector.delete(urls.system_quickmode())
        except ApiError as exc:
            if exc.response is None or exc.response.status_code != 409:
                raise exc

    def set_room_quick_veto(self, room_id: str, quick_veto: QuickVeto) -> None:
        """Set a :class:`~pymultimatic.model.mode.QuickVeto` for a
        :class:`~pymultimatic.model.component.Room`.
        It will override the current
        :class:`~pymultimatic.model.mode.QuickVeto`, if any.

        Args:
            room_id (str): Id of the room.
            quick_veto (QuickVeto): Quick veto to set.
        """
        self._connector.put(urls.room_quick_veto(room_id),
                            payloads.room_quick_veto(
                                self._round(quick_veto.target_temperature),
                                quick_veto.remaining_duration))

    def remove_room_quick_veto(self, room_id: str) -> None:
        """Remove the :class:`~pymultimatic.model.mode.QuickVeto` from a
        :class:`~pymultimatic.model.component.Room`.

        Args:
            room_id (str): Id of the room.
        """
        self._connector.delete(urls.room_quick_veto(room_id))

    def set_zone_quick_veto(self, zone_id: str, quick_veto: QuickVeto) -> None:
        """Set a :class:`~pymultimatic.model.mode.QuickVeto` for a
        :class:`~pymultimatic.model.component.Zone`.
        It will override the current
        :class:`~pymultimatic.model.mode.QuickVeto`, if any.

        Args:
            zone_id (str): Id of the zone.
            quick_veto (QuickVeto): Quick veto to set.
        """
        self._connector.put(urls.zone_quick_veto(zone_id),
                            payloads.zone_quick_veto(
                                self._round(quick_veto.target_temperature)))

    def remove_zone_quick_veto(self, zone_id: str) -> None:
        """Remove the :class:`~pymultimatic.model.mode.QuickVeto` from a
        :class:`~pymultimatic.model.component.Zone`.

        Args:
            zone_id (str): Id of the zone.
        """
        self._connector.delete(urls.zone_quick_veto(zone_id))

    def set_room_setpoint_temperature(self, room_id: str, temperature: float) \
            -> None:
        """Set the new current target temperature for a
        :class:`~pymultimatic.model.component.Room`.

        Note:
            Setting the target temperature this way will actually set the room
            operating mode to
            :class:`~pymultimatic.model.mode.OperatingModes.MANUAL` (and
            target temperature to the desired temperature). It means the target
            temperature will remain the same until you change the target
            temperature again or the operating mode.

        Args:
            room_id (str): Id of the room.
            temperature (float): Target temperature to set.
        """

        _LOGGER.debug("Will try to set room target temperature to %s",
                      temperature)
        self._connector.put(urls.room_temperature_setpoint(room_id),
                            payloads.room_temperature_setpoint(
                                self._round(temperature)))

    def set_zone_setpoint_temperature(self, zone_id: str, temperature: float) \
            -> None:
        """Set the configured temperature for the
        :class:`~pymultimatic.model.mode.SettingModes.DAY` mode.

        Note:
            It won't alter the
            :class:`~pymultimatic.model.mode.OperatingMode`.

        Args:
            zone_id (str): Id of the zone.
            temperature (float): New temperature.
        """
        _LOGGER.debug("Will try to set zone target temperature to %s",
                      temperature)
        self._connector.put(
            urls.zone_heating_setpoint_temperature(zone_id),
            payloads.zone_temperature_setpoint(self._round(temperature)))

    def set_zone_setback_temperature(self, zone_id: str, temperature: float) \
            -> None:
        """Set the configured temperature for the
        :class:`~pymultimatic.model.mode.SettingModes.NIGHT` mode.

        Note:
            It won't alter the
            :class:`~pymultimatic.model.mode.OperatingMode`.

        Args:
            zone_id (str): Id of the zone.
            temperature (float): New temperature.
        """
        _LOGGER.debug("Will try to set zone setback temperature to %s",
                      temperature)
        self._connector.put(urls.zone_heating_setback_temperature(zone_id),
                            payloads.zone_temperature_setback(
                                self._round(temperature)))

    def set_holiday_mode(self, start_date: date, end_date: date,
                         temperature: float) -> None:
        """Set the :class:`~pymultimatic.model.mode.HolidayMode`.

        Args:
            start_date (date): Start date of the holiday mode.
            end_date (date): End date of the holiday mode.
            temperature (float): Target temperature while holiday mode
                :class:`~pymultimatic.model.mode.HolidayMode.is_applied`
        """
        self._connector.put(urls.system_holiday_mode(),
                            payloads.holiday_mode(True, start_date, end_date,
                                                  self._round(temperature)))

    def remove_holiday_mode(self) -> None:
        """Remove :class:`~pymultimatic.model.mode.HolidayMode`.

        Note:
            There is a little workaround here, since the API doesn't simply
            accept a DELETE request to remove the
            :class:`~pymultimatic.model.mode.HolidayMode`, so the manager is
            setting:

            * the start date to two days before

            * the end date to yesterday

            * temperature to frost protection

            * active to `False`

            This will ensure the
            :class:`~pymultimatic.model.mode.HolidayMode` is not active.

        """

        start_date = date.today() - timedelta(days=2)
        end_date = date.today() - timedelta(days=1)
        payload = payloads.holiday_mode(False, start_date, end_date,
                                        constants.FROST_PROTECTION_TEMP)
        self._connector.put(urls.system_holiday_mode(), payload)

    def request_hvac_update(self) -> None:
        """Request an hvac update. This allow the vaillant API to read the data
        from your system.

        This is necessary to update
        :class:`~pymultimatic.model.status.BoilerStatus` and
        :class:`~pymultimatic.model.status.Error`

        Note:
            The **request** done by the manager is done **synchronously**,
            but the **update** processing (basically, reading data from your
            system) is done **asynchronously** by vaillant API.

        Note:
            The manager will try to reduce number of call, I think the call is
            costly. If the
            :class:`~pymultimatic.model.syncstate.SyncState`
            :class:`~pymultimatic.model.syncstate.SyncState.is_pending`,
            the manager will skip the call to the API.


        it can take some times for the update to occur (Most of the
        time, it takes about 1 or 2 minutes before you can see changes, if any)

        It the request is done too often, the API may return an error
        (HTTP 409).

        """

        state = mapper.map_hvac_sync_state(self._connector.get(urls.hvac()))

        if state and not state.is_pending:
            self._connector.put(urls.hvac_update())

    def logout(self) -> None:
        """Get logged out from the API, see
        :func:`~pymultimatic.api.connector.ApiConnector.logout`
        """
        self._connector.logout()

    # pylint: disable=no-self-use
    def _round(self, number: float) -> float:
        """round a float to the nearest 0.5, as vaillant API only accepts 0.5
        step"""
        return round(number * 2) / 2



"""Some file utilities."""
import logging
import os
import pickle
from pathlib import Path
from typing import Any

_LOGGER = logging.getLogger('fileutils')


def load_from_file(path: str, filename: str) -> Any:
    """Load from file."""
    join_path = _join_path(path, filename)
    try:
        with join_path.open("rb") as file:
            return pickle.load(file)

    except (IOError, pickle.PickleError):
        _LOGGER.debug("Cannot open file: %s", join_path, exc_info=True)
        return None


def save_to_file(data: object, path: str, filename: str) -> None:
    """Save to file."""
    join_path = _join_path(path, filename)
    _LOGGER.debug("Will save data to %s", join_path)
    try:
        with join_path.open("wb+") as file:
            pickle.dump(data, file)
    except (IOError, pickle.PickleError):
        _LOGGER.debug("Cannot save data: %s to %s", str(data), join_path,
                      exc_info=True)


def delete_file(path: str, filename: str) -> None:
    """Delete file."""
    join_path = _join_path(path, filename)
    try:
        os.remove(str(join_path))
    except OSError:
        _LOGGER.debug("Cannot delete file %s", join_path, exc_info=True)


def delete_dir(path: str) -> None:
    """Delete directory."""
    try:
        os.rmdir(path)
    except OSError:
        _LOGGER.debug("Cannot delete dir %s", path, exc_info=True)


def _join_path(path: str, filename: str) -> Path:
    """Create path."""
    file_path = Path(path)
    file_path.mkdir(exist_ok=True, parents=True)
    return file_path.joinpath(filename)



"""Groups all API payloads. Payload are always json formatted."""

from datetime import date
from typing import Dict, Optional, Any

_DATE_FORMAT = "%Y-%m-%d"


def hotwater_temperature_setpoint(temperature: float) -> Dict[str, Any]:
    """Payload used to set target temperature for
    :class:`~pymultimatic.model.component.HotWater`.
    """
    return {
        "temperature_setpoint": temperature
    }


def room_temperature_setpoint(temperature: float) -> Dict[str, Any]:
    """Payload used to set target temperature for
    :class:`~pymultimatic.model.component.Room`.
    """
    return {
        "temperatureSetpoint": temperature
    }


def zone_temperature_setpoint(temperature: float) -> Dict[str, Any]:
    """Payload used to set target temperature for
    :class:`~pymultimatic.model.component.Zone`.
    """
    return {
        "setpoint_temperature": temperature
    }


def zone_temperature_setback(temperature: float) -> Dict[str, Any]:
    """Payload used to set setback temperature for
    :class:`~pymultimatic.model.component.Zone`.
    """
    return {
        "setback_temperature": temperature
    }


def hot_water_operating_mode(mode: str) -> Dict[str, Any]:
    """Payload to set operating mode for
    :class:`~pymultimatic.model.component.HotWater`.
    """
    return {"operation_mode": mode}


def room_operating_mode(mode: str) -> Dict[str, Any]:
    """Payload to set operating mode for
    :class:`~pymultimatic.model.component.Room`.
    """
    return {"operationMode": mode}


def zone_operating_mode(mode: str) -> Dict[str, Any]:
    """Payload to set operating mode for
    :class:`~pymultimatic.model.component.Zone`.
    """
    return {"mode": mode}


def quickmode(quick_mode: str, duration: Optional[int] = None) \
        -> Dict[str, Any]:
    """Payload to set :class:`~pymultimatic.model.mode.QuickMode` for the
    system.

    Duration is mandatory (Duration is in minutes, max 1440 =24 hours).
    """
    payload = {
        "quickmode":
            {
                "quickmode": quick_mode
            }
    }

    if duration:
        payload["quickmode"]["duration"] = str(duration)

    return payload


def zone_quick_veto(temperature: float) -> Dict[str, Any]:
    """Payload to set a :class:`~pymultimatic.model.mode.QuickVeto` for a
    :class:`~pymultimatic.model.component.Zone`.

    The duration is not configurable by the API, it's 6 hours
    """
    return {
        "setpoint_temperature": temperature
    }


def room_quick_veto(temperature: float, duration: Optional[int])\
        -> Dict[str, Any]:
    """Payload to set a :class:`~pymultimatic.model.mode.QuickVeto` for a
    :class:`~pymultimatic.model.component.Room`.

    Duration is mandatory (Duration is in minutes, max 1440 =24 hours).
    """

    if not duration:
        duration = 180

    return {
        "temperatureSetpoint": temperature,
        "duration": duration
    }


def holiday_mode(active: bool, start_date: date, end_date: date,
                 temperature: float) -> Dict[str, Any]:
    """Payload to set :class:`~pymultimatic.model.mode.HolidayMode`."""
    return {
        "active": active,
        "start_date": start_date.strftime(_DATE_FORMAT),
        "end_date": end_date.strftime(_DATE_FORMAT),
        "temperature_setpoint": temperature
    }



"""state module"""
from datetime import datetime

import attr

PENDING = 'PENDING'
"""Resource is being updated."""

SYNCED = 'SYNCED'
"""Resource is sync."""

OUTDATED = 'OUTDATED'
"""Resource is outdated."""

INITIALIZING = 'INITIALIZING'
"""Resource is initializing."""


@attr.s
class SyncState:
    """Sync state coming from the API.
    In the vaillant API, most resource you can ask for are flagged with a
    ``resourceState``, with that, you can know if the resource you are asking
    for is synced or not.

    Args:
        state (str): Sate coming from the API.
        timestamp (datetime): state's last update.
        resource_link (str): Link to the resource the state is talking about.
    """

    state = attr.ib(type=str)
    timestamp = attr.ib(type=datetime)
    resource_link = attr.ib(type=str)

    @property
    def is_synced(self) -> bool:
        """bool: Check if state is synced."""
        return self.state == SYNCED

    @property
    def is_pending(self) -> bool:
        """bool: Check if state is pending."""
        return self.state == PENDING

    @property
    def is_outdated(self) -> bool:
        """bool: Check if state is outdated."""
        return self.state == OUTDATED

    @property
    def is_init(self) -> bool:
        """bool: Check if state is initializing."""
        return self.state == INITIALIZING



from setuptools import setup, find_packages

from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='pymultiMATIC',
      version='0.0.10',
      description='Python interface with Vaillant multiMATIC',
      long_description_content_type='text/markdown',
      long_description=long_description,
      url='https://github.com/thomasgermain/pymultiMATIC.git',
      author='Thomas Germain',
      author_email='12560542+thomasgermain@users.noreply.github.com',
      license='MIT',
      packages=find_packages(exclude=
                             ('tests', 'tests/*', '/tests', '/tests/*')),
      zip_safe=False,
      setup_requires=["pytest-runner"],
      install_requires=[
          "requests>=2.20.0,<3.0.0",
          "jsonpickle>=1.0,<2.0",
          "attrs"
      ],
      classifiers=[
          'License :: OSI Approved :: MIT License',
          'Development Status :: 4 - Beta',
          'Programming Language :: Python',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Topic :: Home Automation'
      ]
      )
